from typing import TYPE_CHECKING
import math
import wx
import time

from amulet_map_editor.api.opengl.camera import Projection
from amulet_map_editor.api.opengl.data_types import CameraRotationType

from .base_behaviour import BaseBehaviour
from ..events import (
    InputHeldEvent,
    EVT_INPUT_HELD,
    InputPressEvent,
    EVT_INPUT_PRESS,
    InputReleaseEvent,
    EVT_INPUT_RELEASE,
)
from ..key_config import (
    ACT_MOVE_UP,
    ACT_MOVE_DOWN,
    ACT_MOVE_FORWARDS,
    ACT_MOVE_BACKWARDS,
    ACT_MOVE_LEFT,
    ACT_MOVE_RIGHT,
    ACT_CHANGE_MOUSE_MODE,
    ACT_INCR_SPEED,
    ACT_DECR_SPEED,
    ACT_CHANGE_PROJECTION,
)

if TYPE_CHECKING:
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas


class CameraBehaviour(BaseBehaviour):
    """Adds the normal behaviour for the camera."""

    def __init__(self, canvas: "EditCanvas"):
        super().__init__(canvas)
        self._previous_mouse_lock = self.canvas.camera.rotating = False
        self._toggle_mouse_time = 0
        self._last_camera_rotation: CameraRotationType = (0.0, 0.0)

    def bind_events(self):
        """Set up all events required to run."""
        self.canvas.Bind(EVT_INPUT_HELD, self._on_input_held)
        self.canvas.Bind(EVT_INPUT_PRESS, self._on_input_press)
        self.canvas.Bind(EVT_INPUT_RELEASE, self._on_input_release)
        self.canvas.Bind(wx.EVT_KEY_DOWN, self._on_key_press)
        self.canvas.Bind(wx.EVT_KILL_FOCUS, self._on_loss_focus)
        self.canvas.Bind(wx.EVT_MOTION, self._on_mouse_motion)

    def _on_mouse_motion(self, evt):
        self.canvas.SetFocus()
        evt.Skip()

    def _on_key_press(self, evt: wx.KeyEvent):
        key = evt.GetUnicodeKey() or evt.GetKeyCode()
        if key == wx.WXK_ESCAPE:
            self._escape()
        evt.Skip()

    def _on_input_press(self, evt: InputPressEvent):
        """Logic to run each time the input press event is run."""
        if evt.action_id == ACT_CHANGE_PROJECTION:
            if self.canvas.camera.projection_mode == Projection.PERSPECTIVE:
                self._last_camera_rotation = self.canvas.camera.rotation
                self.canvas.camera.rotation = 180, 90
                self.canvas.camera.projection_mode = Projection.TOP_DOWN
            elif self.canvas.camera.projection_mode == Projection.TOP_DOWN:
                self.canvas.camera.rotation = self._last_camera_rotation
                self.canvas.camera.projection_mode = Projection.PERSPECTIVE
        elif evt.action_id == ACT_CHANGE_MOUSE_MODE:
            self.canvas.SetFocus()
            self._previous_mouse_lock = self.canvas.camera.rotating
            self._capture_mouse()
            self._toggle_mouse_time = time.time()

    def _on_input_release(self, evt: InputReleaseEvent):
        """Logic to run each time the input release event is run."""
        if evt.action_id == ACT_CHANGE_MOUSE_MODE:
            if self._previous_mouse_lock or time.time() - self._toggle_mouse_time > 0.1:
                self._release_mouse()
            else:
                self._capture_mouse()
        elif evt.action_id == ACT_INCR_SPEED:
            if self.canvas.camera.projection_mode == Projection.PERSPECTIVE:
                self.canvas.camera.move_speed *= 1.1
            elif self.canvas.camera.projection_mode == Projection.TOP_DOWN:
                self.canvas.camera.fov = max(0.5, self.canvas.camera.fov / 1.1)
        elif evt.action_id == ACT_DECR_SPEED:
            if self.canvas.camera.projection_mode == Projection.PERSPECTIVE:
                self.canvas.camera.move_speed /= 1.1
            elif self.canvas.camera.projection_mode == Projection.TOP_DOWN:
                self.canvas.camera.fov = min(1000.0, self.canvas.camera.fov * 1.1)

    def _on_input_held(self, evt: InputHeldEvent):
        """Logic to run each time the input held event is run."""
        forward = up = right = pitch = yaw = 0
        up += (ACT_MOVE_UP in evt.action_ids) - (ACT_MOVE_DOWN in evt.action_ids)
        forward += (ACT_MOVE_FORWARDS in evt.action_ids) - (
            ACT_MOVE_BACKWARDS in evt.action_ids
        )
        right += (ACT_MOVE_RIGHT in evt.action_ids) - (ACT_MOVE_LEFT in evt.action_ids)

        if self.canvas.camera.projection_mode == Projection.PERSPECTIVE:
            if self.canvas.camera.rotating:
                pitch = self.canvas.mouse.delta_y * 0.07
                yaw = self.canvas.mouse.delta_x * 0.07
                self.canvas.mouse.warp_middle()
                self.canvas.mouse.reset_delta()
            self.move_camera_relative(forward, up, right, pitch, yaw)
        elif self.canvas.camera.projection_mode == Projection.TOP_DOWN:
            if self.canvas.camera.rotating:
                width, height = self.canvas.GetSize()
                width = max(1, width)
                height = max(1, height)
                forward += (
                    2 * self.canvas.camera.fov * self.canvas.mouse.delta_y / height
                )
                right -= (
                    2
                    * self.canvas.camera.fov
                    * self.canvas.camera.aspect_ratio
                    * self.canvas.mouse.delta_x
                    / width
                )
                self.canvas.mouse.warp_middle()
                self.canvas.mouse.reset_delta()
                self.move_camera_relative(forward, up, right, pitch, yaw)
            else:
                x, y, z = self.canvas.camera.location
                x += right * self.canvas.camera.fov / 30
                z -= forward * self.canvas.camera.fov / 30
                self.canvas.camera.location_rotation = (x, y, z), (180, 90)

        evt.Skip()

    def move_camera_relative(self, forward, up, right, pitch, yaw):
        """Move the camera relative to its current location."""
        if not any((forward, up, right, pitch, yaw)):
            # if not self.canvas.camera.rotating and self._mouse_moved:
            #     self._mouse_moved = False
            #     self._selection_moved = True
            return
        x, y, z = self.canvas.camera.location
        if self.canvas.camera.projection_mode == Projection.PERSPECTIVE:
            ry, rx = self.canvas.camera.rotation
            x += self.canvas.camera.move_speed * -(
                math.cos(math.radians(ry)) * right
                + math.sin(math.radians(ry)) * forward
            )
            y += self.canvas.camera.move_speed * up
            z += self.canvas.camera.move_speed * (
                math.cos(math.radians(ry)) * forward
                - math.sin(math.radians(ry)) * right
            )
            rx += self.canvas.camera.rotate_speed * pitch
            ry += self.canvas.camera.rotate_speed * yaw
        else:
            ry, rx = 180, 90
            x += right
            z -= forward

        self.canvas.camera.location_rotation = (x, y, z), (ry, rx)

    def _capture_mouse(self):
        self.canvas.SetCursor(wx.Cursor(wx.CURSOR_BLANK))
        self.canvas.mouse.warp_middle()
        self.canvas.mouse.reset_delta()
        self.canvas.camera.rotating = True

    def _release_mouse(self):
        """Release the mouse"""
        self.canvas.SetCursor(wx.NullCursor)
        self.canvas.camera.rotating = False

    def _on_loss_focus(self, evt):
        """Event fired when the user tabs out of the window."""
        self._escape()
        evt.Skip()

    def _escape(self):
        """Release the mouse and remove all key presses to the camera doesn't fly off into the distance."""
        # self._persistent_actions.clear()
        self.canvas.buttons.unpress_all()
        self._release_mouse()



from typing import TYPE_CHECKING, Generator, Optional, Any
import numpy
import time

from amulet.api.data_types import Dimension, ChunkCoordinates

from amulet_map_editor import log
from .chunk import RenderChunk
from .region import ChunkManager
from .selection import GreenRenderSelectionGroup
from amulet_map_editor.api.opengl.data_types import (
    CameraLocationType,
    CameraRotationType,
    TransformationMatrix,
)
from amulet_map_editor.api.opengl.resource_pack import (
    OpenGLResourcePackManager,
    OpenGLResourcePack,
)
from amulet_map_editor.api.opengl import Drawable, ThreadedObject, ContextManager

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel


class RenderLevel(OpenGLResourcePackManager, Drawable, ThreadedObject, ContextManager):
    """A RenderLevel holds a reference to a level and manages all the geometry and drawing for that level."""

    def __init__(
        self,
        context_identifier: Any,
        opengl_resource_pack: OpenGLResourcePack,
        level: "BaseLevel",
        draw_box=False,
        draw_floor=False,
        draw_ceil=False,
        limit_bounds=False,
    ):
        """
        Create a new RenderLevel instance.

        :param context_identifier: The identifier for the opengl context.
        :param opengl_resource_pack: The resource pack to use for models and textures.
        :param level: The level to pull data from.
        :param draw_box: Should the box around the level be drawn.
        :param draw_floor: Should the floor below the level be drawn.
        :param draw_ceil: Should the ceiling above the level be drawn.
        :param limit_bounds: Should the chunks be limited to the bounds of the level.
        """
        OpenGLResourcePackManager.__init__(self, opengl_resource_pack)
        ContextManager.__init__(self, context_identifier)
        self._level = level
        self._camera_location: CameraLocationType = (0, 150, 0)
        # yaw (-180 to 180), pitch (-90 to 90)
        self._camera_rotation: CameraRotationType = (0, 90)
        self._dimension: Dimension = level.dimensions[0]
        self._render_distance = 5
        self._garbage_distance = 10
        self._draw_box = draw_box
        self._draw_floor = draw_floor
        self._draw_ceil = draw_ceil
        self._limit_bounds = limit_bounds
        self._selection = GreenRenderSelectionGroup(
            context_identifier, self.resource_pack, self.level.bounds(self.dimension)
        )
        self._chunk_manager = ChunkManager(self.context_identifier, self.resource_pack)

        self._last_rebuild_camera_location: Optional[
            numpy.ndarray
        ] = None  # x, z camera location
        self._needs_rebuild = (
            True  # Should we go back to the beginning and re-find chunks to rebuild
        )
        self._chunk_rebuilds = self._rebuild_generator()
        self._rebuild_time = 0

    @property
    def level(self) -> "BaseLevel":
        return self._level

    @property
    def chunk_manager(self) -> ChunkManager:
        return self._chunk_manager

    def is_closeable(self):
        return True

    def _rebuild_generator(self) -> Generator[Optional[ChunkCoordinates], None, None]:
        """A generator of chunk coordinates to rebuild.
        This is an infinite length generator."""
        while True:
            if self._needs_rebuild:
                self._needs_rebuild = False
                # a set of chunks that are next to chunks that have changed but have not changed themselves
                chunk_rebuilt = set()
                chunk_not_loaded = []  # a list of chunks that have not been loaded

                for chunk_coords in self.chunk_coords():
                    # for all chunks within radius of the player
                    if self.chunk_manager.render_chunk_needs_rebuild(chunk_coords):
                        # if the render chunk exists and the state has changed
                        # rebuild that chunk
                        chunk_rebuilt.add(chunk_coords)
                        yield chunk_coords
                        for offset in [(-1, 0), (1, 0), (0, -1), (0, 1)]:
                            # for all surrounding chunks
                            chunk_coords_ = (
                                chunk_coords[0] + offset[0],
                                chunk_coords[1] + offset[1],
                            )
                            if (
                                chunk_coords_ not in chunk_rebuilt
                                and chunk_coords in self.chunk_manager
                                and not self.chunk_manager.render_chunk_needs_rebuild(
                                    chunk_coords_
                                )
                            ):
                                # if the chunk has not already been rebuilt and it exists
                                # yield it to be rebuilt
                                yield chunk_coords_
                                # store the coords so that it does not get rebuilt twice
                                chunk_rebuilt.add(
                                    chunk_coords_
                                )  # so that it doesn't get picked up again
                        # if the rebuild flag has been set go to the beginning
                        if self._needs_rebuild:
                            break
                    elif chunk_coords not in self.chunk_manager:
                        # if the chunk is not yet loaded, mark it for loading.
                        chunk_not_loaded.append(chunk_coords)
                # if the rebuild flag has been set go to the beginning
                if self._needs_rebuild:
                    continue
                for chunk_coords in chunk_not_loaded:
                    # if the rebuild flag has been set go to the beginning
                    if self._needs_rebuild:
                        break
                    yield chunk_coords
            else:
                yield None

    def thread_action(self):
        # first check if there is a chunk that exists and needs rebuilding
        camera = numpy.asarray(self.camera_location)[[0, 2]]
        if self._last_rebuild_camera_location is None or numpy.sum(
            (self._last_rebuild_camera_location - camera) ** 2
        ) > min(2048, self.render_distance * 16 - 8):
            # if the camera has moved more than 32 blocks set the rebuild flag
            self._needs_rebuild = True
            self._last_rebuild_camera_location = camera

        chunk_coords = next(self._chunk_rebuilds)
        if chunk_coords is not None:
            # generate the chunk
            chunk = RenderChunk(
                self.context_identifier,
                self.resource_pack,
                self.level,
                self.chunk_manager.region_size,
                chunk_coords,
                self.dimension,
                draw_floor=self.draw_floor,
                draw_ceil=self.draw_ceil,
                limit_bounds=self._limit_bounds,
            )

            try:
                chunk.create_geometry()
            except:
                log.error(
                    f"Failed generating chunk geometry for chunk {chunk_coords}",
                    exc_info=True,
                )

            self.chunk_manager.add_render_chunk(chunk)

        t = time.time()
        if t > self._rebuild_time + 1:
            self._rebuild_time = t
            self.chunk_manager.rebuild()

    def enable(self):
        """Enable chunk generation in a new thread."""
        self._needs_rebuild = True

    def unload(self):
        """Unload all loaded data. Can be resumed by calling enable."""
        self.run_garbage_collector(True)

    def close(self):
        self.unload()

    @property
    def camera_location(self) -> CameraLocationType:
        """The x, y, z coordinates of the camera."""
        return self._camera_location

    @camera_location.setter
    def camera_location(self, value: CameraLocationType):
        self._camera_location = value

    @property
    def camera_rotation(self) -> CameraRotationType:
        """The rotation of the camera. (yaw, pitch).
        This should behave the same as how Minecraft handles it.
        """
        return self._camera_rotation

    @camera_rotation.setter
    def camera_rotation(self, value: CameraRotationType):
        """Set the rotation of the camera. (yaw, pitch).
        This should behave the same as how Minecraft handles it.
        """
        self._camera_rotation = value

    @property
    def dimension(self) -> Dimension:
        """The dimension currently being displayed."""
        return self._dimension

    @dimension.setter
    def dimension(self, dimension: Dimension):
        self._dimension = dimension
        self.run_garbage_collector(True)
        self._needs_rebuild = True

    @property
    def render_distance(self) -> int:
        """The radius around the camera within which to load chunks."""
        return self._render_distance

    @render_distance.setter
    def render_distance(self, val: int):
        assert isinstance(val, int), "Render distance must be an int"
        self._render_distance = val
        self._garbage_distance = val + 5
        self._needs_rebuild = True

    @property
    def draw_box(self):
        """Should the selection box around the level be drawn."""
        return self._draw_box

    @property
    def draw_floor(self):
        """Should the floor under the level be drawn."""
        return self._draw_floor

    @property
    def draw_ceil(self):
        """Should the ceiling above the level be drawn."""
        return self._draw_ceil

    def chunk_coords(self) -> Generator[ChunkCoordinates, None, None]:
        """Get all of the chunks to draw/load"""
        # This yield chunk coordinates in a spiral around the camera
        # TODO: Perhaps redesign this to prioritise chunks in front of the camera
        cx, cz = int(self.camera_location[0]) >> 4, int(self.camera_location[2]) >> 4

        sign = 1
        length = 1
        for _ in range(self.render_distance * 2 + 1):
            for _ in range(length):
                yield cx, cz
                cx += sign
            for _ in range(length):
                yield cx, cz
                cz += sign
            sign *= -1
            length += 1

    def draw(self, camera_matrix: TransformationMatrix):
        self._chunk_manager.draw(camera_matrix, self.camera_location)
        if self._draw_box:
            self._selection.draw(
                camera_matrix,
                self.camera_location,
            )

    def run_garbage_collector(self, remove_all=False):
        if remove_all:
            self._chunk_manager.unload()
            self._level.unload()
        else:
            safe_area = (
                self._dimension,
                int(self.camera_location[0] // 16 - self._garbage_distance),
                int(self.camera_location[2] // 16 - self._garbage_distance),
                int(self.camera_location[0] // 16 + self._garbage_distance),
                int(self.camera_location[2] // 16 + self._garbage_distance),
            )
            self._chunk_manager.unload(safe_area[1:])
            self._level.unload(safe_area)

    def _rebuild(self):
        """Unload all the chunks so they can be rebuilt."""
        self._chunk_manager.unload()
        self._needs_rebuild = True

    def rebuild_changed(self):
        """Rebuild the chunks that have changed."""
        self._needs_rebuild = True



import wx
from amulet_map_editor.api.wx.ui.simple import (
    SimpleDialog,
    SimpleScrollablePanel,
    SimpleChoice,
)
from typing import Dict, Tuple, Optional, Union, Sequence

from amulet_map_editor.api.image import ADD_ICON, SUBTRACT_ICON, EDIT_ICON

ModifierKeyType = str
KeyType = Union[int, str]
KeybindGroupIdType = str
KeyActionType = str
ModifierType = Tuple[ModifierKeyType, ...]
SerialisedKeyType = Tuple[ModifierType, KeyType]
KeybindGroup = Dict[KeyActionType, SerialisedKeyType]
ActionLookupType = Dict[SerialisedKeyType, KeyActionType]
KeybindContainer = Dict[KeybindGroupIdType, KeybindGroup]

MouseLeft = "MOUSE_LEFT"
MouseMiddle = "MOUSE_MIDDLE"
MouseRight = "MOUSE_RIGHT"
MouseAux1 = "MOUSE_AUX_1"
MouseAux2 = "MOUSE_AUX_2"
MouseWheelScrollUp = "MOUSE_WHEEL_SCROLL_UP"
MouseWheelScrollDown = "MOUSE_WHEEL_SCROLL_DOWN"
Control = "CTRL"
Shift = "SHIFT"
Alt = "ALT"

Space = "SPACE"
PageUp = "PAGE_UP"
PageDown = "PAGE_DOWN"
Back = "BACK"
Tab = "TAB"
Return = "RETURN"
Escape = "ESCAPE"
Delete = "DELETE"
Start = "START"
Menu = "MENU"
Pause = "PAUSE"
Capital = "CAPITAL"
End = "END"
Home = "HOME"
Left = "LEFT"
Up = "UP"
Right = "RIGHT"
Down = "DOWN"
Select = "SELECT"
Print = "PRINT"
Execute = "EXECUTE"
Snapshot = "SNAPSHOT"
Insert = "INSERT"
Help = "HELP"
Numpad0 = "NUMPAD0"
Numpad1 = "NUMPAD1"
Numpad2 = "NUMPAD2"
Numpad3 = "NUMPAD3"
Numpad4 = "NUMPAD4"
Numpad5 = "NUMPAD5"
Numpad6 = "NUMPAD6"
Numpad7 = "NUMPAD7"
Numpad8 = "NUMPAD8"
Numpad9 = "NUMPAD9"
Multiply = "MULTIPLY"
Add = "ADD"
Separator = "SEPARATOR"
Subtract = "SUBTRACT"
Decimal = "DECIMAL"
Divide = "DIVIDE"
F1 = "F1"
F2 = "F2"
F3 = "F3"
F4 = "F4"
F5 = "F5"
F6 = "F6"
F7 = "F7"
F8 = "F8"
F9 = "F9"
F10 = "F10"
F11 = "F11"
F12 = "F12"
F13 = "F13"
F14 = "F14"
F15 = "F15"
F16 = "F16"
F17 = "F17"
F18 = "F18"
F19 = "F19"
F20 = "F20"
F21 = "F21"
F22 = "F22"
F23 = "F23"
F24 = "F24"
Numlock = "NUMLOCK"
Scroll = "SCROLL"
Numpad_Space = "NUMPAD_SPACE"
Numpad_Tab = "NUMPAD_TAB"
Numpad_Enter = "NUMPAD_ENTER"
Numpad_F1 = "NUMPAD_F1"
Numpad_F2 = "NUMPAD_F2"
Numpad_F3 = "NUMPAD_F3"
Numpad_F4 = "NUMPAD_F4"
Numpad_Home = "NUMPAD_HOME"
Numpad_Left = "NUMPAD_LEFT"
Numpad_Up = "NUMPAD_UP"
Numpad_Right = "NUMPAD_RIGHT"
Numpad_Down = "NUMPAD_DOWN"
Numpad_Pageup = "NUMPAD_PAGEUP"
Numpad_Pagedown = "NUMPAD_PAGEDOWN"
Numpad_End = "NUMPAD_END"
Numpad_Begin = "NUMPAD_BEGIN"
Numpad_Insert = "NUMPAD_INSERT"
Numpad_Delete = "NUMPAD_DELETE"
Numpad_Equal = "NUMPAD_EQUAL"
Numpad_Multiply = "NUMPAD_MULTIPLY"
Numpad_Add = "NUMPAD_ADD"
Numpad_Separator = "NUMPAD_SEPARATOR"
Numpad_Subtract = "NUMPAD_SUBTRACT"
Numpad_Decimal = "NUMPAD_DECIMAL"
Numpad_Divide = "NUMPAD_DIVIDE"

key_string_map = {
    wx.WXK_CONTROL: Control,
    wx.WXK_SHIFT: Shift,
    wx.WXK_ALT: Alt,
    wx.WXK_SPACE: Space,
    wx.WXK_PAGEUP: PageUp,
    wx.WXK_PAGEDOWN: PageDown,
    wx.WXK_BACK: Back,
    wx.WXK_TAB: Tab,
    wx.WXK_RETURN: Return,
    wx.WXK_ESCAPE: Escape,
    wx.WXK_DELETE: Delete,
    wx.WXK_START: Start,
    wx.WXK_MENU: Menu,
    wx.WXK_PAUSE: Pause,
    wx.WXK_CAPITAL: Capital,
    wx.WXK_END: End,
    wx.WXK_HOME: Home,
    wx.WXK_LEFT: Left,
    wx.WXK_UP: Up,
    wx.WXK_RIGHT: Right,
    wx.WXK_DOWN: Down,
    wx.WXK_SELECT: Select,
    wx.WXK_PRINT: Print,
    wx.WXK_EXECUTE: Execute,
    wx.WXK_SNAPSHOT: Snapshot,
    wx.WXK_INSERT: Insert,
    wx.WXK_HELP: Help,
    wx.WXK_NUMPAD0: Numpad0,
    wx.WXK_NUMPAD1: Numpad1,
    wx.WXK_NUMPAD2: Numpad2,
    wx.WXK_NUMPAD3: Numpad3,
    wx.WXK_NUMPAD4: Numpad4,
    wx.WXK_NUMPAD5: Numpad5,
    wx.WXK_NUMPAD6: Numpad6,
    wx.WXK_NUMPAD7: Numpad7,
    wx.WXK_NUMPAD8: Numpad8,
    wx.WXK_NUMPAD9: Numpad9,
    wx.WXK_MULTIPLY: Multiply,
    wx.WXK_ADD: Add,
    wx.WXK_SEPARATOR: Separator,
    wx.WXK_SUBTRACT: Subtract,
    wx.WXK_DECIMAL: Decimal,
    wx.WXK_DIVIDE: Divide,
    wx.WXK_F1: F1,
    wx.WXK_F2: F2,
    wx.WXK_F3: F3,
    wx.WXK_F4: F4,
    wx.WXK_F5: F5,
    wx.WXK_F6: F6,
    wx.WXK_F7: F7,
    wx.WXK_F8: F8,
    wx.WXK_F9: F9,
    wx.WXK_F10: F10,
    wx.WXK_F11: F11,
    wx.WXK_F12: F12,
    wx.WXK_F13: F13,
    wx.WXK_F14: F14,
    wx.WXK_F15: F15,
    wx.WXK_F16: F16,
    wx.WXK_F17: F17,
    wx.WXK_F18: F18,
    wx.WXK_F19: F19,
    wx.WXK_F20: F20,
    wx.WXK_F21: F21,
    wx.WXK_F22: F22,
    wx.WXK_F23: F23,
    wx.WXK_F24: F24,
    wx.WXK_NUMLOCK: Numlock,
    wx.WXK_SCROLL: Scroll,
    wx.WXK_NUMPAD_SPACE: Numpad_Space,
    wx.WXK_NUMPAD_TAB: Numpad_Tab,
    wx.WXK_NUMPAD_ENTER: Numpad_Enter,
    wx.WXK_NUMPAD_F1: Numpad_F1,
    wx.WXK_NUMPAD_F2: Numpad_F2,
    wx.WXK_NUMPAD_F3: Numpad_F3,
    wx.WXK_NUMPAD_F4: Numpad_F4,
    wx.WXK_NUMPAD_HOME: Numpad_Home,
    wx.WXK_NUMPAD_LEFT: Numpad_Left,
    wx.WXK_NUMPAD_UP: Numpad_Up,
    wx.WXK_NUMPAD_RIGHT: Numpad_Right,
    wx.WXK_NUMPAD_DOWN: Numpad_Down,
    wx.WXK_NUMPAD_PAGEUP: Numpad_Pageup,
    wx.WXK_NUMPAD_PAGEDOWN: Numpad_Pagedown,
    wx.WXK_NUMPAD_END: Numpad_End,
    wx.WXK_NUMPAD_BEGIN: Numpad_Begin,
    wx.WXK_NUMPAD_INSERT: Numpad_Insert,
    wx.WXK_NUMPAD_DELETE: Numpad_Delete,
    wx.WXK_NUMPAD_EQUAL: Numpad_Equal,
    wx.WXK_NUMPAD_MULTIPLY: Numpad_Multiply,
    wx.WXK_NUMPAD_ADD: Numpad_Add,
    wx.WXK_NUMPAD_SEPARATOR: Numpad_Separator,
    wx.WXK_NUMPAD_SUBTRACT: Numpad_Subtract,
    wx.WXK_NUMPAD_DECIMAL: Numpad_Decimal,
    wx.WXK_NUMPAD_DIVIDE: Numpad_Divide,
}

_mouse_events = {
    wx.EVT_LEFT_DOWN.evtType[0]: MouseLeft,
    wx.EVT_LEFT_UP.evtType[0]: MouseLeft,
    wx.EVT_MIDDLE_DOWN.evtType[0]: MouseMiddle,
    wx.EVT_MIDDLE_UP.evtType[0]: MouseMiddle,
    wx.EVT_RIGHT_DOWN.evtType[0]: MouseRight,
    wx.EVT_RIGHT_UP.evtType[0]: MouseRight,
    wx.EVT_MOUSE_AUX1_DOWN.evtType[0]: MouseAux1,
    wx.EVT_MOUSE_AUX1_UP.evtType[0]: MouseAux1,
    wx.EVT_MOUSE_AUX2_DOWN.evtType[0]: MouseAux2,
    wx.EVT_MOUSE_AUX2_UP.evtType[0]: MouseAux2,
}


def serialise_modifier(
    evt: Union[wx.KeyEvent, wx.MouseEvent], key: int
) -> ModifierType:

    modifier = []
    if evt.ControlDown():
        if key not in (wx.WXK_SHIFT, wx.WXK_ALT):
            # if control is pressed the real key must not be a modifier
            modifier.append(Control)
            if evt.ShiftDown():
                modifier.append(Shift)
            if evt.AltDown():
                modifier.append(Alt)
    return tuple(modifier)


def serialise_key(evt: Union[wx.KeyEvent, wx.MouseEvent]) -> Optional[KeyType]:
    """Get the serialised version of the key that was pressed/released."""
    if isinstance(evt, wx.KeyEvent):
        key = evt.GetUnicodeKey() or evt.GetKeyCode()

        if 33 <= key <= 126:
            key = chr(key).upper()
        elif key in key_string_map:
            key = key_string_map[key]
        else:
            key = f"UNKNOWN KEY {key}"
        return key
    elif isinstance(evt, wx.MouseEvent):
        key = evt.GetEventType()
        if key in wx.EVT_MOUSEWHEEL.evtType:
            if evt.GetWheelRotation() < 0:
                return MouseWheelScrollDown
            elif evt.GetWheelRotation() > 0:
                return MouseWheelScrollUp
        elif key in _mouse_events:
            return _mouse_events[key]


def serialise_key_event(
    evt: Union[wx.KeyEvent, wx.MouseEvent]
) -> Optional[SerialisedKeyType]:
    if isinstance(evt, wx.KeyEvent):

        key = evt.GetUnicodeKey() or evt.GetKeyCode()
        if key == wx.WXK_CONTROL:
            return
        modifier = serialise_modifier(evt, key)

        if 33 <= key <= 126:
            key = chr(key).upper()
        elif key in key_string_map:
            key = key_string_map[key]
        else:
            key = f"UNKNOWN KEY {key}"
        return modifier, key
    elif isinstance(evt, wx.MouseEvent):
        key = evt.GetEventType()
        modifier = serialise_modifier(evt, key)
        if key in wx.EVT_MOUSEWHEEL.evtType:
            if evt.GetWheelRotation() < 0:
                return modifier, MouseWheelScrollDown
            elif evt.GetWheelRotation() > 0:
                return modifier, MouseWheelScrollUp
        elif key in _mouse_events:
            return modifier, _mouse_events[key]


def stringify_key(key: SerialisedKeyType) -> str:
    return " + ".join([str(s) for s in key[0] + (key[1],)])


class KeyCatcher(wx.Dialog):
    def __init__(self, parent: wx.Window, action: str):
        super().__init__(
            parent,
            title=f"Press the key you want assigned to {action}",
            style=wx.DEFAULT_DIALOG_STYLE | wx.WANTS_CHARS,
        )

        self._key = ((), "NONE")

        self.Bind(wx.EVT_LEFT_DOWN, self._on_key)
        self.Bind(wx.EVT_MIDDLE_DOWN, self._on_key)
        self.Bind(wx.EVT_RIGHT_DOWN, self._on_key)
        self.Bind(wx.EVT_KEY_DOWN, self._on_key)
        self.Bind(wx.EVT_MOUSEWHEEL, self._on_key)
        self.Bind(wx.EVT_MOUSE_AUX1_DOWN, self._on_key)
        self.Bind(wx.EVT_MOUSE_AUX2_DOWN, self._on_key)

    def _on_key(self, evt):
        key = serialise_key_event(evt)
        if key is not None:
            self._key = key
            self.EndModal(1)

    @property
    def key(self) -> SerialisedKeyType:
        return self._key


# TODO: make any key able to be a modifier. Instead of registering keys on press register them on release.
#  When a key is pressed store it to a set of persistent keys. When a key is released that is the triggering
#  key and all persistent keys are modifiers.
#  In the actual program detect all keys on press/release as normal but also store the set of persistent keys.
#  When serialising give the key that was pressed/released along with all the modifiers.
#  Return all actions that use the triggering key as the triggering key and use a sub-set of the persistent keys.


class KeyConfigDialog(SimpleDialog):
    def __init__(
        self,
        parent: wx.Window,
        selected_group: KeybindGroupIdType,
        entries: Sequence[KeyActionType],
        fixed_keybinds: KeybindContainer,
        user_keybinds: KeybindContainer,
    ):
        super().__init__(parent, "Key Select")
        self._key_config = KeyConfig(
            self, selected_group, entries, fixed_keybinds, user_keybinds
        )
        self.sizer.Add(self._key_config, 1, wx.EXPAND)
        self.Layout()
        self.Fit()

    @property
    def options(self) -> Tuple[KeybindContainer, KeybindGroupIdType, KeybindGroup]:
        return self._key_config.options


class KeyConfig(wx.BoxSizer):
    def __init__(
        self,
        parent: wx.Window,
        selected_group: KeybindGroupIdType,
        entries: Sequence[KeyActionType],
        fixed_keybinds: KeybindContainer,
        user_keybinds: KeybindContainer,
    ):
        super().__init__(wx.VERTICAL)
        self._entries = entries
        self._fixed_keybinds = fixed_keybinds
        self._user_keybinds = user_keybinds

        top_sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.Add(top_sizer, 0, wx.EXPAND)
        self._choice = SimpleChoice(
            parent,
            list(self._fixed_keybinds.keys()) + list(self._user_keybinds.keys()),
            selected_group,
        )
        self._choice.Bind(wx.EVT_CHOICE, self._on_group_change)
        top_sizer.Add(self._choice, 1, wx.ALL | wx.EXPAND, 5)

        add = wx.BitmapButton(parent, bitmap=ADD_ICON.bitmap(32, 32))
        add.Bind(wx.EVT_BUTTON, lambda evt: self._create_new_group())
        top_sizer.Add(add, 0, wx.ALL, 5)

        self._delete = wx.BitmapButton(parent, bitmap=SUBTRACT_ICON.bitmap(32, 32))
        self._delete.Bind(wx.EVT_BUTTON, lambda evt: self._delete_group())
        top_sizer.Add(self._delete, 0, wx.ALL, 5)

        self._rename = wx.BitmapButton(parent, bitmap=EDIT_ICON.bitmap(32, 32))
        self._rename.Bind(wx.EVT_BUTTON, lambda evt: self._rename_group())
        top_sizer.Add(self._rename, 0, wx.ALL, 5)

        self._options = SimpleScrollablePanel(parent, size=(500, 500))
        self.Add(self._options, 1, wx.EXPAND)

        grid_sizer = wx.GridSizer(len(entries), 2, 5, 5)
        self._options.sizer.Add(grid_sizer, 0, wx.ALL | wx.EXPAND, 5)
        self._key_buttons: Dict[str, wx.Button] = {}
        for action in entries:
            grid_sizer.Add(
                wx.StaticText(self._options, label=action.title()), 0, wx.ALIGN_CENTER
            )
            self._key_buttons[action] = button = wx.Button(self._options)
            button.Bind(wx.EVT_BUTTON, lambda evt, a=action: self._modify_button(a))
            grid_sizer.Add(button, 0, wx.EXPAND)
        self._rebuild_buttons()

    def _rebuild_buttons(self):
        group_id = self._choice.GetCurrentString()
        if group_id in self._fixed_keybinds:
            group = self._fixed_keybinds[group_id]
            self._delete.Disable()
            self._rename.Disable()
        else:
            group = self._user_keybinds[group_id]
            self._delete.Enable()
            self._rename.Enable()

        for action in self._entries:
            self._key_buttons[action].SetLabel(
                stringify_key(group.get(action, ((), "NONE")))
            )

    def _rebuild_choice(self, group_name=None):
        index = self._choice.GetSelection()
        self._choice.SetItems(
            list(self._fixed_keybinds.keys()) + list(self._user_keybinds.keys())
        )
        if group_name is not None:
            self._choice.SetSelection(self._choice.FindString(group_name))
        else:
            self._choice.SetSelection(max(index - 1, 0))
        self._rebuild_buttons()

    def _delete_group(self):
        group = self._choice.GetCurrentString()
        if group in self._user_keybinds:
            del self._user_keybinds[group]
            self._rebuild_choice()

    def _request_group_name(self) -> Optional[str]:
        group_name = ""
        while group_name == "":
            msg = wx.TextEntryDialog(self._options, "Enter a new group name.")
            if msg.ShowModal() == wx.ID_OK:
                group_name = msg.GetValue()
                if (
                    group_name in self._fixed_keybinds
                    or group_name in self._user_keybinds
                ):
                    group_name = ""
            else:
                return
        return group_name

    def _rename_group(self):
        old_group_name = self._choice.GetCurrentString()
        if old_group_name in self._user_keybinds:
            group_name = self._request_group_name()
            if group_name is None:
                return
            group = self._user_keybinds[old_group_name]
            del self._user_keybinds[old_group_name]
            self._user_keybinds[group_name] = group
            self._rebuild_choice(group_name)

    def _create_new_group(self):
        group_name = self._request_group_name()
        if group_name is None:
            return
        old_group_name = self._choice.GetCurrentString()
        if old_group_name in self._fixed_keybinds:
            group = self._fixed_keybinds[old_group_name]
        else:
            group = self._user_keybinds[old_group_name]
        self._user_keybinds[group_name] = group.copy()
        self._rebuild_choice(group_name)

    def _modify_button(self, action):
        if self._choice.GetCurrentString() in self._fixed_keybinds:
            msg = wx.MessageDialog(
                self._options,
                "The active key group is not editable. Would you like to create a new group to edit?",
                style=wx.YES_NO,
            )
            if msg.ShowModal() == wx.ID_YES:
                self._create_new_group()
            else:
                return
        group_name = self._choice.GetCurrentString()
        if group_name in self._user_keybinds:
            catcher = KeyCatcher(self._options, action)
            catcher.ShowModal()
            self._user_keybinds[group_name][action] = catcher.key
            self._rebuild_buttons()

    def _on_group_change(self, evt):
        self._rebuild_buttons()

    @property
    def options(self) -> Tuple[KeybindContainer, KeybindGroupIdType, KeybindGroup]:
        keybind_group = self._choice.GetCurrentString()
        if keybind_group in self._fixed_keybinds:
            keybinds = self._fixed_keybinds[keybind_group]
        else:
            keybinds = self._user_keybinds[keybind_group]
        return self._user_keybinds, keybind_group, keybinds



from amulet_map_editor.api.datatypes import MenuData


class BaseProgram:
    def enable(self):
        """Run when the panel is shown/enabled"""
        pass

    def disable(self):
        """Run when the panel is hidden/disabled"""
        pass

    def is_closeable(self) -> bool:
        """
        Check if it is safe to close the UI.
        If this is going to return False it should notify the user.
        :return: True if the program can be closed, False otherwise
        """
        return True

    def close(self):
        """Fully close the UI. Called when destroying the UI."""
        pass

    def menu(self, menu: MenuData) -> MenuData:
        return menu



from wx.lib import newevent

from .camera import (
    CameraMovedEvent,
    EVT_CAMERA_MOVED,
    ProjectionChangedEvent,
    EVT_PROJECTION_CHANGED,
)

PreDrawEvent, EVT_PRE_DRAW = newevent.NewEvent()
DrawEvent, EVT_DRAW = newevent.NewEvent()
PostDrawEvent, EVT_POST_DRAW = newevent.NewEvent()



from typing import Optional
import time
from concurrent.futures import ThreadPoolExecutor, Future

from amulet_map_editor.api.opengl import ThreadedObjectContainer

ThreadingEnabled = True
if ThreadingEnabled:

    class ChunkGenerator(ThreadedObjectContainer, ThreadPoolExecutor):
        def __init__(self):
            ThreadedObjectContainer.__init__(self)
            ThreadPoolExecutor.__init__(self, max_workers=1)
            self._enabled = False
            self._generator: Optional[Future] = None

        def start(self):
            if not self._enabled:
                self._enabled = True
                self._generator = self.submit(self._generate_chunks)

        def stop(self):
            if self._enabled:
                self._enabled = False
                self._generator.result()

        def _generate_chunks(self):
            while self._enabled:
                start_time = time.time()
                while time.time() - start_time < 1 / 120:
                    self.thread_action()
                time.sleep(1 / 60)


else:

    class ChunkGenerator(ThreadedObjectContainer):
        def start(self):
            pass

        def stop(self):
            pass



import wx
import wx.lib.scrolledpanel
from typing import List

import PyMCTranslate

from amulet_map_editor.api.image import (
    UP_CARET,
    DOWN_CARET,
    TRASH,
    ADD_ICON,
    MAXIMIZE,
    MINIMIZE,
)
from amulet_map_editor.api.wx.ui.block_select import BlockDefine, EVT_PROPERTIES_CHANGE


class MultiBlockDefine(wx.lib.scrolledpanel.ScrolledPanel):
    def __init__(self, parent, translation_manager, style=0):
        super().__init__(parent, style=style)
        self.SetupScrolling()

        self._translation_manager = translation_manager

        self._sizer = wx.BoxSizer(wx.VERTICAL)

        self._add_button = wx.BitmapButton(self, bitmap=ADD_ICON.bitmap(18, 18))
        self._sizer.Add(self._add_button, 0, wx.TOP | wx.LEFT | wx.RIGHT | wx.EXPAND, 5)

        self._block_picker_sizer = wx.BoxSizer(wx.VERTICAL)

        self._block_picker_sizer.Add(
            _CollapsibleBlockDefine(self, translation_manager), 0, wx.TOP | wx.EXPAND, 5
        )
        self._sizer.Add(
            self._block_picker_sizer, 0, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.BOTTOM, 5
        )

        self.SetSizerAndFit(self._sizer)
        self.Layout()

        self._add_button.Bind(wx.EVT_BUTTON, self._add)
        self._fix_enabled_buttons()

    def _add(self, evt):
        self.Freeze()
        self.collapse()
        block_picker = _CollapsibleBlockDefine(self, self._translation_manager)
        self._block_picker_sizer.Add(block_picker, 1, wx.TOP | wx.EXPAND, 5)
        self._block_picker_sizer.Layout()
        self._sizer.Layout()
        self.Layout()
        self._fix_enabled_buttons()
        self.Refresh()
        self.Thaw()

    def move_up(self, obj):
        sizer = self._block_picker_sizer
        index = [child.Window for child in sizer.GetChildren()].index(obj)

        sizer.Detach(obj)
        sizer.Insert(index - 1 if index > 0 else 0, obj, 0, wx.TOP | wx.EXPAND, 5)
        self._fix_enabled_buttons()
        self.Layout()

    def move_down(self, obj):
        sizer = self._block_picker_sizer
        length = sizer.ItemCount
        index = [child.Window for child in sizer.GetChildren()].index(obj)

        sizer.Detach(obj)
        sizer.Insert(
            index + 1 if index < length - 1 else length - 1,
            obj,
            0,
            wx.TOP | wx.EXPAND,
            5,
        )
        self._fix_enabled_buttons()
        self.Layout()

    def delete(self, obj):
        obj.Hide()
        obj.Destroy()
        self._fix_enabled_buttons()
        self.Layout()

    def collapse(self):
        for child in self._block_picker_sizer.GetChildren():
            child.Window.collapsed = True

    def _fix_enabled_buttons(self):
        windows: List[_CollapsibleBlockDefine] = [
            child.Window for child in self._block_picker_sizer.GetChildren()
        ]
        for window in windows:
            window.up_button.Enable()
            window.down_button.Enable()
            window.delete_button.Enable()

        if len(windows) >= 1:
            windows[0].up_button.Disable()
            windows[-1].down_button.Disable()
        if len(windows) == 1:
            windows[0].delete_button.Disable()


class _CollapsibleBlockDefine(wx.Panel):
    def __init__(self, parent: MultiBlockDefine, translation_manager, collapsed=False):
        super().__init__(parent, style=wx.BORDER_SIMPLE)

        self.EXPAND = MAXIMIZE.bitmap(18, 18)
        self.COLLAPSE = MINIMIZE.bitmap(18, 18)

        self._collapsed = collapsed

        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)

        header_sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(header_sizer, 0, wx.ALL, 5)

        self.expand_button = wx.BitmapButton(self, bitmap=self.EXPAND)
        header_sizer.Add(self.expand_button, 0, 5)

        self.up_button = wx.BitmapButton(self, bitmap=UP_CARET.bitmap(18, 18))
        header_sizer.Add(self.up_button, 0, wx.LEFT, 5)
        self.up_button.Bind(wx.EVT_BUTTON, lambda evt: parent.move_up(self))

        self.down_button = wx.BitmapButton(self, bitmap=DOWN_CARET.bitmap(18, 18))
        header_sizer.Add(self.down_button, 0, wx.LEFT, 5)
        self.down_button.Bind(wx.EVT_BUTTON, lambda evt: parent.move_down(self))

        self.delete_button = wx.BitmapButton(self, bitmap=TRASH.bitmap(18, 18))
        header_sizer.Add(self.delete_button, 0, wx.LEFT, 5)
        self.delete_button.Bind(wx.EVT_BUTTON, lambda evt: parent.delete(self))

        self.block_define = BlockDefine(self, translation_manager, wx.HORIZONTAL)
        sizer.Add(self.block_define, 1, wx.LEFT | wx.RIGHT | wx.BOTTOM | wx.EXPAND, 5)
        self.collapsed = collapsed

        self.block_label = wx.StaticText(
            self,
            label=self._gen_block_string(),
            style=wx.ST_ELLIPSIZE_END | wx.ST_NO_AUTORESIZE,
            size=(500, -1),
        )
        header_sizer.Add(self.block_label, 1, wx.ALIGN_CENTER_VERTICAL | wx.LEFT, 5)

        self.expand_button.Bind(
            wx.EVT_BUTTON, lambda evt: self._toggle_block_expand(parent)
        )
        self.block_define.Bind(EVT_PROPERTIES_CHANGE, self._on_properties_change)

    @property
    def collapsed(self) -> bool:
        return self._collapsed

    @collapsed.setter
    def collapsed(self, collapsed: bool):
        self._collapsed = collapsed
        if self._collapsed:
            self.expand_button.SetBitmap(self.EXPAND)
            self.block_define.Hide()
        else:
            self.expand_button.SetBitmap(self.COLLAPSE)
            self.block_define.Show()
        self.TopLevelParent.Layout()

    def _toggle_block_expand(self, parent: MultiBlockDefine):
        if self.collapsed:
            parent.collapse()
        self.collapsed = not self.collapsed

    def _on_properties_change(self, evt):
        self.block_label.SetLabel(self._gen_block_string())
        self.TopLevelParent.Layout()
        evt.Skip()

    def _gen_block_string(self):
        base = f"{self.block_define.namespace}:{self.block_define.block_name}"
        properties = ",".join(
            (
                f"{key}={value}"
                for key, value in self.block_define.str_properties.items()
            )
        )
        return f"{base}[{properties}]" if properties else base


if __name__ == "__main__":

    def main():
        app = wx.App()
        translation_manager = PyMCTranslate.new_translation_manager()
        dialog = wx.Dialog(None, style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)
        sizer = wx.BoxSizer()
        dialog.SetSizer(sizer)
        sizer.Add(MultiBlockDefine(dialog, translation_manager), 1, wx.EXPAND)
        dialog.Show()
        dialog.Fit()
        app.MainLoop()

    main()



class ContextManager:
    """Store the uuid of the context this data applies to."""

    def __init__(self, context_identifier: str):
        self._context_identifier = context_identifier

    @property
    def context_identifier(self) -> str:
        return self._context_identifier



from typing import Optional, Tuple
import wx
import re

from amulet.api.data_types import PointCoordinates

from amulet_map_editor import lang
from amulet_map_editor.api.wx.ui.simple import SimpleDialog
from amulet_map_editor.api import image

CoordRegex = re.compile(
    r"^"  # match the start
    r"\s*"  # leading whitespace
    r"(?P<x>-?[0-9]+\.?[0-9]*)"  # the x coordinate
    r"((?:,\s*)|(?:\s+))"  # separator 1
    r"(?P<y>-?[0-9]+\.?[0-9]*)"  # the y coordinate
    r"((?:,\s*)|(?:\s+))"  # separator 2
    r"(?P<z>-?[0-9]+\.?[0-9]*)"  # the z coordinate
    r",?\s*"  # trailing comma and whitespace
    r"$"  # matches the end
)


def show_goto(
    parent: wx.Window, x: float, y: float, z: float
) -> Optional[Tuple[float, float, float]]:
    dialog = GoTo(parent, lang.get("program_3d_edit.goto_ui.title"), (x, y, z))
    if dialog.ShowModal() == wx.ID_OK:
        return dialog.location


class GoTo(SimpleDialog):
    def __init__(self, parent: wx.Window, title: str, start: PointCoordinates):
        super().__init__(parent, title, wx.HORIZONTAL)
        x, y, z = start
        x_text = wx.StaticText(self, label=lang.get("program_3d_edit.goto_ui.x_label"))
        self.x = wx.SpinCtrlDouble(self, min=-30000000, max=30000000, initial=x)
        self.x.SetToolTip(lang.get("program_3d_edit.goto_ui.x_label_tooltip"))
        self.x.SetDigits(2)
        y_text = wx.StaticText(self, label=lang.get("program_3d_edit.goto_ui.y_label"))
        self.y = wx.SpinCtrlDouble(self, min=-30000000, max=30000000, initial=y)
        self.y.SetToolTip(lang.get("program_3d_edit.goto_ui.y_label_tooltip"))
        self.y.SetDigits(2)
        z_text = wx.StaticText(self, label=lang.get("program_3d_edit.goto_ui.z_label"))
        self.z = wx.SpinCtrlDouble(self, min=-30000000, max=30000000, initial=z)
        self.z.SetToolTip(lang.get("program_3d_edit.goto_ui.z_label_tooltip"))
        self.z.SetDigits(2)
        self.sizer.Add(x_text, 0, wx.CENTER | wx.ALL, 5)
        self.sizer.Add(self.x, 1, wx.CENTER | wx.TOP | wx.BOTTOM | wx.RIGHT, 5)
        self.sizer.Add(y_text, 0, wx.CENTER | wx.TOP | wx.BOTTOM | wx.RIGHT, 5)
        self.sizer.Add(self.y, 1, wx.CENTER | wx.TOP | wx.BOTTOM | wx.RIGHT, 5)
        self.sizer.Add(z_text, 0, wx.CENTER | wx.TOP | wx.BOTTOM | wx.RIGHT, 5)
        self.sizer.Add(self.z, 1, wx.CENTER | wx.TOP | wx.BOTTOM | wx.RIGHT, 5)
        self.x.Bind(wx.EVT_CHAR, self._on_text)
        self.y.Bind(wx.EVT_CHAR, self._on_text)
        self.z.Bind(wx.EVT_CHAR, self._on_text)

        copy_button = wx.BitmapButton(
            self, bitmap=image.icon.tablericons.copy.bitmap(20, 20)
        )
        copy_button.Bind(wx.EVT_BUTTON, lambda evt: self._copy())
        copy_button.SetToolTip(lang.get("program_3d_edit.goto_ui.copy_button_tooltip"))
        self.bottom_sizer.Insert(0, copy_button, 0, wx.ALL, 5)

        paste_button = wx.BitmapButton(
            self, bitmap=image.icon.tablericons.clipboard.bitmap(20, 20)
        )
        paste_button.Bind(wx.EVT_BUTTON, lambda evt: self._paste())
        paste_button.SetToolTip(
            lang.get("program_3d_edit.goto_ui.paste_button_tooltip")
        )
        self.bottom_sizer.Insert(1, paste_button, 0, wx.TOP | wx.BOTTOM | wx.RIGHT, 5)
        self.Fit()

    @property
    def location(self) -> PointCoordinates:
        return self.x.GetValue(), self.y.GetValue(), self.z.GetValue()

    def _on_text(self, evt):
        if evt.ControlDown() and evt.GetKeyCode() == 3:
            # Ctrl+C
            self._copy()
        elif evt.ControlDown() and evt.GetKeyCode() == 22:
            # Ctrl+V
            if not self._paste():
                evt.Skip()
        else:
            evt.Skip()

    def _copy(self):
        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(
                wx.TextDataObject(
                    "{} {} {}".format(
                        round(self.x.GetValue(), 5),
                        round(self.y.GetValue(), 5),
                        round(self.z.GetValue(), 5),
                    )
                )
            )
            wx.TheClipboard.Close()

    def _paste(self) -> bool:
        text = ""
        text_data = wx.TextDataObject()
        if wx.TheClipboard.Open():
            success = wx.TheClipboard.GetData(text_data)
            wx.TheClipboard.Close()
            if success:
                text = text_data.GetText()
        match = CoordRegex.fullmatch(text)
        if match:
            self.x.SetValue(float(match.group("x")))
            self.y.SetValue(float(match.group("y")))
            self.z.SetValue(float(match.group("z")))
            return True
        return False



import wx
from wx import glcanvas
from OpenGL.GL import (
    GL_DEPTH_TEST,
    glEnable,
    glGenTextures,
    GL_CULL_FACE,
    glDepthFunc,
    GL_LEQUAL,
    GL_BLEND,
    glBlendFunc,
    GL_SRC_ALPHA,
    GL_ONE_MINUS_SRC_ALPHA,
    glDeleteTextures,
)

import uuid
import sys


class BaseCanvas(glcanvas.GLCanvas):
    def __init__(self, parent: wx.Window):
        display_attributes = wx.glcanvas.GLAttributes()
        display_attributes.PlatformDefaults().MinRGBA(8, 8, 8, 8).DoubleBuffer().Depth(
            24
        ).EndList()
        super().__init__(
            parent,
            display_attributes,
            size=parent.GetClientSize(),
            style=wx.WANTS_CHARS,
        )

        # create a UUID for the context. Used to get shaders
        self._context_identifier = str(uuid.uuid4())

        if sys.platform == "linux":
            # setup the OpenGL context. This apparently fixes #84
            self._context = glcanvas.GLContext(self)
        else:
            context_attributes = wx.glcanvas.GLContextAttrs()
            context_attributes.CoreProfile().Robust().ResetIsolation().EndList()
            self._context = glcanvas.GLContext(
                self, ctxAttrs=context_attributes
            )  # setup the OpenGL context
        self.SetCurrent(self._context)
        self._gl_texture_atlas = glGenTextures(1)  # Create the atlas texture location
        self._setup_opengl()  # set some OpenGL states

    @property
    def context_identifier(self) -> str:
        return self._context_identifier

    def _setup_opengl(self):
        glEnable(GL_DEPTH_TEST)
        glEnable(GL_CULL_FACE)
        glDepthFunc(GL_LEQUAL)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

    def close(self):
        glDeleteTextures([self._gl_texture_atlas])



from typing import TYPE_CHECKING, Tuple
import numpy
import time

from amulet.api.selection import SelectionGroup, SelectionBox
from .pointer_behaviour import PointerBehaviour
from amulet_map_editor.api.opengl.mesh.selection import (
    RenderSelectionGroup,
    RenderSelection,
)
from ..events import (
    InputPressEvent,
    EVT_INPUT_PRESS,
    InputReleaseEvent,
    EVT_INPUT_RELEASE,
    EVT_SELECTION_CHANGE,
)
from amulet_map_editor.api.opengl.camera import Projection
from ..key_config import (
    ACT_BOX_CLICK,
    ACT_DESELECT_ALL_BOXES,
    ACT_DELESECT_BOX,
)

if TYPE_CHECKING:
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas


class ChunkSelectionBehaviour(PointerBehaviour):
    """Adds the behaviour for a chunk based selection."""

    def __init__(self, canvas: "EditCanvas"):
        super().__init__(canvas)
        self._editing_selection = RenderSelection(
            self.canvas.context_identifier, self.canvas.renderer.opengl_resource_pack
        )
        self._selection = RenderSelectionGroup(
            self.canvas.context_identifier,
            self.canvas.renderer.opengl_resource_pack,
        )
        self._editing = False
        self._press_time = 0
        self._start_box = numpy.zeros((2, 3))

    def bind_events(self):
        super().bind_events()
        self.canvas.Bind(EVT_INPUT_PRESS, self._on_input_press)
        self.canvas.Bind(EVT_INPUT_RELEASE, self._on_input_release)
        self.canvas.Bind(EVT_SELECTION_CHANGE, self._on_selection_change)

    def _create_undo(self):
        """Write the current state to the global selection triggering an undo point."""
        self.canvas.selection.selection_group = self._selection.selection_group

    def _chunkify_selection(self):
        selections = []
        for box in self.canvas.selection.selection_group.selection_boxes:
            min_point = (
                numpy.floor(box.min_array / self.canvas.world.sub_chunk_size)
                * self.canvas.world.sub_chunk_size
            )
            max_point = (
                numpy.ceil(box.max_array / self.canvas.world.sub_chunk_size)
                * self.canvas.world.sub_chunk_size
            )
            bounds = self.canvas.world.bounds(self.canvas.dimension)
            min_point[1] = bounds.min[1]
            max_point[1] = bounds.max[1]
            selections.append(SelectionBox(min_point, max_point))
        selection_group = SelectionGroup(selections)
        if selection_group != self.canvas.selection.selection_group:
            # if the above code modified the selection
            self.canvas.selection.selection_group = selection_group
            # this will indirectly update the renderer by updating the global selection
        elif selection_group != self._selection.selection_group:
            # if the above code did not change the selection but it does not match the renderer
            self._selection.selection_group = selection_group

    def enable(self):
        self._chunkify_selection()
        self._editing = False

    def _on_selection_change(self, evt):
        self._chunkify_selection()
        evt.Skip()

    def _on_input_press(self, evt: InputPressEvent):
        if evt.action_id == ACT_BOX_CLICK:
            if not self._editing:
                self._press_time = time.time()
                self._editing = True
                self._start_box = self._pointer.bounds
                self._update_pointer()
        elif evt.action_id == ACT_DESELECT_ALL_BOXES:
            self._selection.selection_group = SelectionGroup()
            self._create_undo()
        elif evt.action_id == ACT_DELESECT_BOX:
            if ACT_DESELECT_ALL_BOXES not in self.canvas.buttons.pressed_actions:
                self._selection.selection_group = SelectionGroup()
                self._create_undo()
        evt.Skip()

    def _get_editing_selection(self) -> Tuple[numpy.ndarray, numpy.ndarray]:
        """Get the minimum and maximum points of the editing selection.
        This is based on the stored starting pointer and the current pointer."""
        points = numpy.concatenate([self._start_box, self._pointer.bounds], 0)
        return numpy.min(points, 0), numpy.max(points, 0)

    def _on_input_release(self, evt: InputReleaseEvent):
        if evt.action_id == ACT_BOX_CLICK:
            if self._editing and time.time() - self._press_time > 0.1:
                self._editing = False
                box = SelectionBox(*self._get_editing_selection())
                if SelectionGroup(box).is_subset(self._selection.selection_group):
                    # subtract
                    self._selection.selection_group = (
                        self._selection.selection_group.subtract(box)
                    )
                else:
                    self._selection.selection_group = (
                        self._selection.selection_group.union(box)
                    )
                self._create_undo()
        evt.Skip()

    def _update_pointer(self):
        """Update the pointer and selection location."""
        if self.canvas.camera.projection_mode == Projection.TOP_DOWN:
            x, z = self.get_2d_mouse_location()
            location = numpy.array([x, 0, z])
        else:
            camera_location = self.canvas.camera.location
            look_vector = self.look_vector()
            (
                box,
                max_distance,
            ) = self._selection.selection_group.closest_vector_intersection(
                camera_location, look_vector
            )
            location, hit = self.closest_block_3d(min(max_distance, 100))

        chunk_size = self.canvas.world.sub_chunk_size
        location_min: numpy.ndarray = (location // chunk_size) * chunk_size
        location_max = location_min + chunk_size
        bounds = self.canvas.world.bounds(self.canvas.dimension)
        location_min[1] = bounds.min[1]
        location_max[1] = bounds.max[1]
        self._pointer.point1, self._pointer.point2 = location_min, location_max
        if self._editing:
            (
                self._editing_selection.point1,
                self._editing_selection.point2,
            ) = self._get_editing_selection()

    def draw(self):
        if self.canvas.camera.projection_mode == Projection.TOP_DOWN:
            camera = None
        else:
            camera = self.canvas.camera.location
        self._selection.draw(self.canvas.camera.transformation_matrix, camera)
        if self._editing:
            self._editing_selection.draw(
                self.canvas.camera.transformation_matrix, camera
            )
        super().draw()



import wx
from typing import TYPE_CHECKING

from amulet.api.selection import SelectionGroup
from amulet.api.data_types import Dimension, OperationReturnType

from amulet_map_editor.programs.edit.api.operations import DefaultOperationUI

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas


class SimpleOperationPanel(wx.Panel, DefaultOperationUI):
    def __init__(
        self,
        parent: wx.Window,
        canvas: "EditCanvas",
        world: "BaseLevel",
        options_path: str,
    ):
        wx.Panel.__init__(self, parent)
        DefaultOperationUI.__init__(self, parent, canvas, world, options_path)

        self._sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self._sizer)

    def _add_run_button(self, label="Run Operation"):
        self._run_button = wx.Button(self, label=label)
        self._run_button.Bind(wx.EVT_BUTTON, self._run_operation)
        self._sizer.Add(self._run_button, 0, wx.ALL | wx.ALIGN_CENTRE_HORIZONTAL, 5)
        self.Layout()

    def _run_operation(self, _):
        self.canvas.run_operation(
            lambda: self._operation(
                self.world, self.canvas.dimension, self.canvas.selection.selection_group
            )
        )

    def _operation(
        self, world: "BaseLevel", dimension: Dimension, selection: SelectionGroup
    ) -> OperationReturnType:
        raise NotImplementedError



from typing import TYPE_CHECKING

from amulet.operations.fill import fill
from amulet.api.selection import SelectionGroup
from amulet.api.block import UniversalAirBlock
from amulet.api.data_types import Dimension, OperationReturnType

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel


def delete(
    world: "BaseLevel", dimension: Dimension, selection: SelectionGroup
) -> OperationReturnType:
    yield from fill(
        world,
        dimension,
        selection,
        UniversalAirBlock,
    )



from typing import TYPE_CHECKING
from amulet_map_editor.api.opengl.mesh.selection import RenderSelectionGroup
from ..events import EVT_SELECTION_CHANGE

from .base_behaviour import BaseBehaviour
from amulet_map_editor.api.opengl.camera import Projection

if TYPE_CHECKING:
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas


class StaticSelectionBehaviour(BaseBehaviour):
    """Adds the logic for a static selection."""

    def __init__(self, canvas: "EditCanvas"):
        super().__init__(canvas)
        self._selection = RenderSelectionGroup(
            self.canvas.context_identifier,
            self.canvas.renderer.opengl_resource_pack,
        )

    def bind_events(self):
        self.canvas.Bind(EVT_SELECTION_CHANGE, self._on_selection_change)

    def _on_selection_change(self, evt):
        """Update the render selection based on the updated selection."""
        self.update_selection()
        evt.Skip()

    def update_selection(self):
        """Pull the latest selection from the canvas."""
        self._selection.selection_group = self.canvas.selection.selection_group

    def draw(self):
        if self.canvas.camera.projection_mode == Projection.TOP_DOWN:
            camera = None
        else:
            camera = self.canvas.camera.location
        self._selection.draw(self.canvas.camera.transformation_matrix, camera)



from __future__ import annotations

import atexit

import wx

from amulet_map_editor import CONFIG

PRE_EXISTING_CONFIG = CONFIG.get("window_preferences", {})


def write_config():
    CONFIG.put("window_preferences", PRE_EXISTING_CONFIG)


atexit.register(write_config)


def on_idle(self):
    global PRE_EXISTING_CONFIG

    qualified_name = ".".join((self.__module__, self.__class__.__name__))

    def wrapper(evt):
        update_cfg = False
        if self.__resized:
            self.__resized = False
            update_cfg = True
        if self.__moved:
            self.__moved = False
            update_cfg = True
        if update_cfg:
            PRE_EXISTING_CONFIG[qualified_name] = {
                "size": self.GetSize().Get(),
                "position": self.GetPosition().Get(),
            }
            self.Refresh()
            self.Layout()
        evt.Skip()

    return wrapper


def on_size(self):
    def wrapper(evt):
        self.__resized = True
        evt.Skip()

    return wrapper


def on_move(self):
    def wrapper(evt):
        self.__moved = True
        evt.Skip()

    return wrapper


def preserve_ui_preferences(clazz):
    original_init = clazz.__init__
    qualified_name = ".".join((clazz.__module__, clazz.__name__))

    def __init__(self, *args, **kwargs):
        original_init(self, *args, **kwargs)
        self.__resized = False
        self.__moved = False

        if qualified_name in PRE_EXISTING_CONFIG:
            self.SetSize(PRE_EXISTING_CONFIG[qualified_name]["size"])
            self.SetPosition(PRE_EXISTING_CONFIG[qualified_name]["position"])
            self.Refresh()

        self.Bind(wx.EVT_MOVE, on_move(self))
        self.Bind(wx.EVT_SIZE, on_size(self))
        self.Bind(wx.EVT_IDLE, on_idle(self))

    clazz.__init__ = __init__
    return clazz



from typing import Tuple, Dict
from types import ModuleType
import os
import traceback
import pkgutil
import importlib
import re

from .loader import (
    BaseOperationLoader,
    OperationLoader,
    UIOperationLoader,
)
from .util import STOCK_PLUGINS_DIR, STOCK_PLUGINS_NAME, CUSTOM_PLUGINS_DIR, load_module

import amulet_map_editor
from amulet_map_editor import log


class BaseOperationManager:
    OperationClass: BaseOperationLoader = None

    def __init__(self, group_name: str):
        """Set up an operation manager that handles operations for a set group.

        :param group_name: The name of the group. This is used to pick the directory.
        """
        # The loaded operations. Stored under their identifier.
        assert issubclass(self.OperationClass, BaseOperationLoader)
        self._group_name = group_name
        self._operations: Dict[str, BaseOperationLoader] = {}
        self.reload()

    @property
    def operations(self) -> Tuple[BaseOperationLoader, ...]:
        return tuple(self._operations.values())

    def get_operation(self, operation_id: str):
        return self._operations[operation_id]

    def __getitem__(self, operation_id: str):
        return self._operations[operation_id]

    def reload(self):
        """Unload the old operations and reload them all.
        If new operations are created, load them.
        If old operations were removed, remove them."""
        self._operations.clear()
        self._load_pyinstaller(f"{STOCK_PLUGINS_NAME}.{self._group_name}")
        self._load_dir(os.path.join(STOCK_PLUGINS_DIR, self._group_name))
        self._load_dir(os.path.join(CUSTOM_PLUGINS_DIR, self._group_name), True)

    def _load_dir(self, path: str, make: bool = False):
        """Load all operations in a set directory."""
        if make:
            os.makedirs(path, exist_ok=True)
        if os.path.isdir(path):
            for obj_name in os.listdir(path):
                if obj_name in {"__init__.py", "__pycache__"}:
                    continue
                obj_path = os.path.join(path, obj_name)
                if os.path.isfile(obj_path) and not obj_path.endswith(".py"):
                    continue
                try:
                    mod = load_module(obj_path)
                except ImportError:
                    log.warning(
                        f"Failed to import {obj_path}.\n{traceback.format_exc()}"
                    )
                else:
                    self._load_module(obj_path, mod)

    def _load_pyinstaller(self, package: str):
        # pyinstaller support
        toc = set()
        for importer in pkgutil.iter_importers(amulet_map_editor.__name__):
            if hasattr(importer, "toc"):
                toc |= importer.toc
        prefix = f"{package}."
        match = re.compile(f"^{re.escape(prefix)}[a-zA-Z0-9_]*$")
        for module_name in toc:
            if match.fullmatch(module_name):
                try:
                    mod = importlib.import_module(module_name)
                except ImportError:
                    log.warning(
                        f"Failed to import {module_name}.\n{traceback.format_exc()}"
                    )
                else:
                    self._load_module(module_name, mod)

    def _load_module(self, obj_path: str, mod: ModuleType):
        if hasattr(mod, "export"):
            export = mod.export
            if isinstance(export, dict):
                self._load_operation(obj_path, export)
            elif isinstance(export, (list, tuple)):
                for i, export_dict in export:
                    self._load_operation(f"{obj_path}[{i}]", export_dict)
            else:
                log.error(f"The format of export in {obj_path} is invalid.")
        else:
            log.error(f"export is not present in {obj_path}")

    def _load_operation(self, identifier: str, export_dict: dict):
        """Create an instance of a subclass of BaseOperationLoader to load the operation.
        When finished and the class is valid store it in self._operations

        :param identifier: A unique identifier for this operation. <path>[0]
        :param export_dict: The dictionary defining the operation.
        :return:
        """
        op = self.OperationClass(identifier, export_dict)
        if op.is_valid:
            self._operations[op.identifier] = op


class OperationManager(BaseOperationManager):
    OperationClass = OperationLoader
    """A class to manage a group of function operations that run immediately when called."""


class UIOperationManager(BaseOperationManager):
    OperationClass = UIOperationLoader
    """A class to manage a group of UI operations that return a wx UI when called."""



from typing import List
from amulet_map_editor.api.wx.util.key_config import (
    KeybindContainer,
    KeybindGroup,
    KeybindGroupIdType,
    KeyActionType,
    Space,
    Shift,
    MouseLeft,
    MouseRight,
    MouseWheelScrollUp,
    MouseWheelScrollDown,
    Control,
    Alt,
    Tab,
)

ACT_MOVE_UP = "ACT_MOVE_UP"
ACT_MOVE_DOWN = "ACT_MOVE_DOWN"
ACT_MOVE_FORWARDS = "ACT_MOVE_FORWARDS"
ACT_MOVE_BACKWARDS = "ACT_MOVE_BACKWARDS"
ACT_MOVE_LEFT = "ACT_MOVE_LEFT"
ACT_MOVE_RIGHT = "ACT_MOVE_RIGHT"
ACT_BOX_CLICK = "ACT_BOX_CLICK"
ACT_BOX_CLICK_ADD = "ACT_BOX_CLICK_ADD"
ACT_CHANGE_MOUSE_MODE = "ACT_CHANGE_MOUSE_MODE"
ACT_INCR_SPEED = "ACT_INCR_SPEED"
ACT_DECR_SPEED = "ACT_DECR_SPEED"
ACT_INCR_SELECT_DISTANCE = "ACT_INCR_SELECT_DISTANCE"
ACT_DECR_SELECT_DISTANCE = "ACT_DECR_SELECT_DISTANCE"
ACT_DESELECT_ALL_BOXES = "ACT_DESELECT_ALL_BOXES"
ACT_DELESECT_BOX = "ACT_DELESECT_BOX"
ACT_INSPECT_BLOCK = "ACT_INSPECT_BLOCK"
ACT_CHANGE_PROJECTION = "ACT_CHANGE_PROJECTION"

KeybindKeys: List[KeyActionType] = [
    ACT_MOVE_UP,
    ACT_MOVE_DOWN,
    ACT_MOVE_FORWARDS,
    ACT_MOVE_BACKWARDS,
    ACT_MOVE_LEFT,
    ACT_MOVE_RIGHT,
    ACT_BOX_CLICK,
    ACT_BOX_CLICK_ADD,
    ACT_CHANGE_MOUSE_MODE,
    ACT_INCR_SPEED,
    ACT_DECR_SPEED,
    ACT_INCR_SELECT_DISTANCE,
    ACT_DECR_SELECT_DISTANCE,
    ACT_DESELECT_ALL_BOXES,
    ACT_DELESECT_BOX,
    ACT_INSPECT_BLOCK,
    ACT_CHANGE_PROJECTION,
]

PresetKeybinds: KeybindContainer = {
    "right": {
        ACT_MOVE_UP: ((), Space),
        ACT_MOVE_DOWN: ((), Shift),
        ACT_MOVE_FORWARDS: ((), "W"),
        ACT_MOVE_BACKWARDS: ((), "S"),
        ACT_MOVE_LEFT: ((), "A"),
        ACT_MOVE_RIGHT: ((), "D"),
        ACT_BOX_CLICK: ((), MouseLeft),
        ACT_BOX_CLICK_ADD: ((Control,), MouseLeft),
        ACT_CHANGE_MOUSE_MODE: ((), MouseRight),
        ACT_INCR_SPEED: ((), MouseWheelScrollUp),
        ACT_DECR_SPEED: ((), MouseWheelScrollDown),
        ACT_INCR_SELECT_DISTANCE: ((), "R"),
        ACT_DECR_SELECT_DISTANCE: ((), "F"),
        ACT_DESELECT_ALL_BOXES: ((Control, Shift), "D"),
        ACT_DELESECT_BOX: ((Control,), "D"),
        ACT_INSPECT_BLOCK: ((), Alt),
        ACT_CHANGE_PROJECTION: ((), Tab),
    },
    "right_laptop": {
        ACT_MOVE_UP: ((), Space),
        ACT_MOVE_DOWN: ((), Shift),
        ACT_MOVE_FORWARDS: ((), "W"),
        ACT_MOVE_BACKWARDS: ((), "S"),
        ACT_MOVE_LEFT: ((), "A"),
        ACT_MOVE_RIGHT: ((), "D"),
        ACT_BOX_CLICK: ((), MouseLeft),
        ACT_BOX_CLICK_ADD: ((Control,), MouseLeft),
        ACT_CHANGE_MOUSE_MODE: ((), MouseRight),
        ACT_INCR_SPEED: ((), "."),
        ACT_DECR_SPEED: ((), ","),
        ACT_INCR_SELECT_DISTANCE: ((), "R"),
        ACT_DECR_SELECT_DISTANCE: ((), "F"),
        ACT_DESELECT_ALL_BOXES: ((Control, Shift), "D"),
        ACT_DELESECT_BOX: ((Control,), "D"),
        ACT_INSPECT_BLOCK: ((), Alt),
        ACT_CHANGE_PROJECTION: ((), Tab),
    },
    "left": {
        ACT_MOVE_UP: ((), Space),
        ACT_MOVE_DOWN: ((), ";"),
        ACT_MOVE_FORWARDS: ((), "I"),
        ACT_MOVE_BACKWARDS: ((), "K"),
        ACT_MOVE_LEFT: ((), "J"),
        ACT_MOVE_RIGHT: ((), "L"),
        ACT_BOX_CLICK: ((), MouseLeft),
        ACT_BOX_CLICK_ADD: ((Control,), MouseLeft),
        ACT_CHANGE_MOUSE_MODE: ((), MouseRight),
        ACT_INCR_SPEED: ((), MouseWheelScrollUp),
        ACT_DECR_SPEED: ((), MouseWheelScrollDown),
        ACT_INCR_SELECT_DISTANCE: ((), "Y"),
        ACT_DECR_SELECT_DISTANCE: ((), "H"),
        ACT_DESELECT_ALL_BOXES: ((Control, Shift), "D"),
        ACT_DELESECT_BOX: ((Control,), "D"),
        ACT_INSPECT_BLOCK: ((), Alt),
        ACT_CHANGE_PROJECTION: ((), Tab),
    },
    "left_laptop": {
        ACT_MOVE_UP: ((), Space),
        ACT_MOVE_DOWN: ((), ";"),
        ACT_MOVE_FORWARDS: ((), "I"),
        ACT_MOVE_BACKWARDS: ((), "K"),
        ACT_MOVE_LEFT: ((), "J"),
        ACT_MOVE_RIGHT: ((), "L"),
        ACT_BOX_CLICK: ((), MouseLeft),
        ACT_BOX_CLICK_ADD: ((Control,), MouseLeft),
        ACT_CHANGE_MOUSE_MODE: ((), MouseRight),
        ACT_INCR_SPEED: ((), "."),
        ACT_DECR_SPEED: ((), ","),
        ACT_INCR_SELECT_DISTANCE: ((), "Y"),
        ACT_DECR_SELECT_DISTANCE: ((), "H"),
        ACT_DESELECT_ALL_BOXES: ((Control, Shift), "D"),
        ACT_DELESECT_BOX: ((Control,), "D"),
        ACT_INSPECT_BLOCK: ((), Alt),
        ACT_CHANGE_PROJECTION: ((), Tab),
    },
}

DefaultKeybindGroupId: KeybindGroupIdType = "right"
DefaultKeys: KeybindGroup = PresetKeybinds[DefaultKeybindGroupId]



from wx.lib import newevent

from amulet_map_editor.api.opengl.camera import (
    CameraMovedEvent,
    EVT_CAMERA_MOVED,
    ProjectionChangedEvent,
    EVT_PROJECTION_CHANGED,
)

from amulet_map_editor.api.opengl.events import (
    PreDrawEvent,
    EVT_PRE_DRAW,
    DrawEvent,
    EVT_DRAW,
    PostDrawEvent,
    EVT_POST_DRAW,
)

from amulet_map_editor.api.wx.util.button_input import (
    InputPressEvent,
    InputHeldEvent,
    InputReleaseEvent,
    EVT_INPUT_PRESS,
    EVT_INPUT_HELD,
    EVT_INPUT_RELEASE,
)

from .selection import SelectionChangeEvent, EVT_SELECTION_CHANGE


DimensionChangeEvent, EVT_DIMENSION_CHANGE = newevent.NewEvent()

# the active tool changed
ToolChangeEvent, EVT_TOOL_CHANGE = newevent.NewEvent()
PasteEvent, EVT_PASTE = newevent.NewEvent()

UndoEvent, EVT_UNDO = newevent.NewEvent()
RedoEvent, EVT_REDO = newevent.NewEvent()
CreateUndoEvent, EVT_CREATE_UNDO = newevent.NewEvent()
SaveEvent, EVT_SAVE = newevent.NewEvent()
EditCloseEvent, EVT_EDIT_CLOSE = newevent.NewEvent()



from typing import TYPE_CHECKING, Tuple, Optional
import time
import numpy
import wx
from wx.lib import newevent

from ..events import (
    InputPressEvent,
    InputReleaseEvent,
    EVT_INPUT_RELEASE,
    EVT_SELECTION_CHANGE,
)
from amulet.api.selection import SelectionGroup, SelectionBox
from amulet.api.data_types import (
    PointCoordinates,
    BlockCoordinates,
)
from amulet_map_editor.api.opengl.camera import Projection
from amulet_map_editor.api.opengl.mesh.selection import (
    RenderSelectionEditable,
    RenderSelectionGroupHighlightable,
)

from .pointer_behaviour import PointerBehaviour
from ..key_config import (
    ACT_BOX_CLICK,
    ACT_BOX_CLICK_ADD,
    ACT_INCR_SELECT_DISTANCE,
    ACT_DECR_SELECT_DISTANCE,
    ACT_DESELECT_ALL_BOXES,
    ACT_DELESECT_BOX,
)

if TYPE_CHECKING:
    from ..canvas import EditCanvas

# the active selection was changed from local code. Not fired by active_block_positions.setter
_RenderBoxChangeEvent = wx.NewEventType()
EVT_RENDER_BOX_CHANGE = wx.PyEventBinder(_RenderBoxChangeEvent)


class RenderBoxChangeEvent(wx.PyEvent):
    """Run when the selection is changed by third party code."""

    def __init__(self, point1: BlockCoordinates, point2: BlockCoordinates):
        wx.PyEvent.__init__(self, eventType=_RenderBoxChangeEvent)
        self._point1 = point1
        self._point2 = point2

    @property
    def point1(self) -> BlockCoordinates:
        return self._point1

    @property
    def point2(self) -> BlockCoordinates:
        return self._point2

    @property
    def points(self) -> Tuple[BlockCoordinates, BlockCoordinates]:
        return self._point1, self._point2


# run when self._editing is set to True or the active selection is removed. Locks external editing
RenderBoxDisableInputsEvent, EVT_RENDER_BOX_DISABLE_INPUTS = newevent.NewEvent()

# run when self._editing is set to False and there is an active selection. Allows external editing
RenderBoxEnableInputsEvent, EVT_RENDER_BOX_ENABLE_INPUTS = newevent.NewEvent()

NPArray2x3 = numpy.ndarray
NPVector3 = numpy.ndarray


class BlockSelectionBehaviour(PointerBehaviour):
    """Adds the behaviour for a block based selection."""

    def __init__(self, canvas: "EditCanvas"):
        super().__init__(canvas)
        self._selection: RenderSelectionGroupHighlightable = (
            RenderSelectionGroupHighlightable(
                self.canvas.context_identifier,
                self.canvas.renderer.opengl_resource_pack,
            )
        )

        self._active_selection: Optional[RenderSelectionEditable] = None
        self._editing = False  # is the active selection being created or resized
        self._press_time = 0  # the time when the last box edit started
        self._start_point_1: NPArray2x3 = numpy.zeros(
            (2, 3)
        )  # the state of the cursor when editing starts
        self._start_point_2: NPArray2x3 = numpy.zeros(
            (2, 3)
        )  # the state of the cursor when editing starts
        self._highlight = False  # is a box being highlighted
        self._initial_box: Optional[
            NPArray2x3
        ] = None  # the state of the box when editing started
        self._pointer_mask: NPArray2x3 = numpy.zeros((2, 3), dtype=bool)
        self._resizing = False  # is a box being resized
        self._pointer_distance2 = 0  # the pointer distance used when resizing

    def _create_active_selection(self):
        """Create the active selection if it does not exist."""
        if self._active_selection is None:
            self._active_selection = RenderSelectionEditable(
                self.canvas.context_identifier,
                self.canvas.renderer.opengl_resource_pack,
            )

    def _unload_active_selection(self):
        """Unload the active selection if it exists."""
        if self._active_selection is not None:
            self._active_selection.unload()
            self._active_selection = None

    def bind_events(self):
        super().bind_events()
        self.canvas.Bind(EVT_INPUT_RELEASE, self._on_input_release)
        self.canvas.Bind(EVT_SELECTION_CHANGE, self._on_selection_change)
        self.canvas.Bind(wx.EVT_KEY_DOWN, self._on_key_press)

    def enable(self):
        self.pull_selection()

    def _on_selection_change(self, evt):
        self.pull_selection()
        evt.Skip()

    def push_selection(self):
        """Write the current state to the global selection triggering an undo point."""
        self.canvas.selection.selection_group = self.selection_group

    def pull_selection(self):
        """Pull the selection from the canvas."""
        selection = self.canvas.selection.selection_group
        self.selection_group = selection
        if selection:
            self._enable_inputs()
        else:
            self._disable_inputs()

    def _post_change_event(self):
        super()._post_change_event()
        wx.PostEvent(self.canvas, RenderBoxChangeEvent(*self.active_block_positions))

    def _disable_inputs(self):
        wx.PostEvent(self.canvas, RenderBoxDisableInputsEvent())

    def _enable_inputs(self):
        wx.PostEvent(self.canvas, RenderBoxEnableInputsEvent())

    def _on_input_press(self, evt: InputPressEvent):
        if evt.action_id == ACT_INCR_SELECT_DISTANCE:
            if self._resizing:
                self._pointer_distance2 += 1
            else:
                self._pointer_distance += 1
            self._pointer_moved = True
        elif evt.action_id == ACT_DECR_SELECT_DISTANCE:
            if self._resizing:
                self._pointer_distance2 -= 1
            else:
                self._pointer_distance -= 1
            self._pointer_moved = True
        elif evt.action_id == ACT_BOX_CLICK:
            if not self._editing:
                default_create = True
                self._press_time = time.time()
                self._disable_inputs()

                if ACT_BOX_CLICK_ADD in self.canvas.buttons.pressed_actions:
                    # create a new box
                    if self._active_selection is not None:
                        # move the existing active to the inactive
                        self._selection.selection_group = (
                            self._selection.selection_group
                            + self._active_selection.selection_group
                        )

                elif self._highlight:
                    # edit the highlighted box
                    box_index, distance, faces = self._get_box_faces()
                    if box_index is None:
                        # this shouldn't happen but just to be safe.
                        self._selection.selection_group = SelectionGroup()
                    else:
                        if box_index < len(self._selection):
                            # a static selection box is highlighted
                            selection_group = self.selection_group
                            self._selection.selection_group = (
                                selection_group[:box_index]
                                + selection_group[box_index + 1 :]
                            )
                            self._active_selection.selection_box = selection_group[
                                box_index
                            ]
                            self._post_change_event()

                        (
                            self._start_point_1,
                            self._start_point_2,
                        ) = self._get_active_points()
                        self._pointer_distance2 = distance + 0.01
                        self._pointer_mask = faces
                        self._active_selection.set_highlight_edges(faces)
                        self._initial_box = self._active_selection.points
                        self._resizing = True
                        default_create = False
                        self._pointer_moved = True
                else:
                    # clear the selection
                    self._selection.selection_group = SelectionGroup()

                self._create_active_selection()
                self._active_selection.locked = False
                self._editing = True
                if default_create:
                    self._start_point_1 = self._pointer.bounds
                    self._pointer_mask = numpy.array(
                        [[False] * 3, [True] * 3], dtype=bool
                    )
                    self._active_selection.set_highlight_edges(self._pointer_mask)
                    self._initial_box = None
                    self._resizing = False

        elif evt.action_id == ACT_DESELECT_ALL_BOXES:
            if self.selection_group:
                self.selection_group = SelectionGroup()
                self.push_selection()
                self._disable_inputs()
                self._post_change_event()
            else:
                self._escape()
        elif evt.action_id == ACT_DELESECT_BOX:
            if ACT_DESELECT_ALL_BOXES not in self.canvas.buttons.pressed_actions:
                selection_group = self.selection_group
                if selection_group:
                    self.selection_group = selection_group[:-1]
                    self.push_selection()
                    if self._active_selection is None:
                        self._disable_inputs()
                        self._post_change_event()
                else:
                    self._escape()
        evt.Skip()

    def _on_key_press(self, evt: wx.KeyEvent):
        key = evt.GetUnicodeKey() or evt.GetKeyCode()
        if key == wx.WXK_ESCAPE:
            self._escape()
        evt.Skip()

    def _on_input_release(self, evt: InputReleaseEvent):
        if evt.action_id == ACT_BOX_CLICK:
            if self._editing and time.time() - self._press_time > 0.1:
                self._editing = self._resizing = False
                self._enable_inputs()
                self._active_selection.locked = True
                self.push_selection()
        evt.Skip()

    def _escape(self):
        """Reset the state to how it was before editing."""
        if self._editing:
            if self._initial_box is None:
                # there was no initial box
                if self._selection:
                    # static selection exists. Load the last into the active.
                    self._create_active_selection()
                    selection_group = self.selection_group
                    self._selection.selection_group = selection_group[:-1]
                    self._active_selection.selection_box = selection_group[-1]
                    self._active_selection.locked = True
                    self._enable_inputs()
                else:
                    self._unload_active_selection()
            else:
                # there was an initial box
                self._create_active_selection()
                self._active_selection.points = self._initial_box
                self._active_selection.locked = True
                self._enable_inputs()
            self._editing = self._resizing = False
            self._highlight = False
            self._post_change_event()

    def _get_active_points(self) -> Tuple[NPArray2x3, NPArray2x3]:
        """Get the 1x1x1 box coords for the active selection."""
        p1, p2 = numpy.zeros((2, 3)), numpy.zeros((2, 3))
        if self._active_selection is not None:
            p1[0], p2[0] = self._active_selection.points
            p1[1] = p1[0] + 1
            p2[1] = p2[0] + 1
            mult = p1[0] < p2[0]
            p1 -= 1 * numpy.logical_not(mult)
            p2 -= 1 * mult
        return p1, p2

    @property
    def active_block_positions(
        self,
    ) -> Tuple[Tuple[int, int, int], Tuple[int, int, int]]:
        """Get the active box positions.
        The coordinates for the maximum point of the box will be one less because this is the block position."""
        if self._active_selection is None:
            return (0, 0, 0), (0, 0, 0)
        else:
            p1, p2 = self._get_active_points()
            return tuple(p1[0].tolist()), tuple(p2[0].tolist())

    @active_block_positions.setter
    def active_block_positions(
        self, positions: Tuple[Tuple[int, int, int], Tuple[int, int, int]]
    ):
        """Set the active box positions.
        This should only be used when not editing.
        The coordinates for the maximum point of the box will be one greater because this is the block position."""
        if self._active_selection is not None and not self._editing:
            self._pointer_mask[:] = False
            self._start_point_1[:] = positions[0]
            self._start_point_2[:] = positions[1]
            self._start_point_1[1] += 1
            self._start_point_2[1] += 1
            (
                self._active_selection.point1,
                self._active_selection.point2,
            ) = self._get_editing_selection()
            self.push_selection()
        else:
            self._disable_inputs()

    @property
    def selection_group(self) -> SelectionGroup:
        """The selection group of the static and active selection.
        If the active selection is being resized it will not be included in this."""
        selection_group = self._selection.selection_group
        if self._active_selection is not None and not self._editing:
            selection_group += self._active_selection.selection_group
        return selection_group

    @selection_group.setter
    def selection_group(self, selection_group: SelectionGroup):
        """Set the selection group of the static and active selection.
        This will only trigger a grapical change and will not update the global selection.
        A call to push_selection is required to push the updated selection to the global selection."""
        self._escape()
        if len(selection_group) == 0:
            # unload the active selection
            if self._active_selection is not None:
                self._unload_active_selection()

        if len(selection_group) <= 1:
            # unload the selection group
            self._selection.selection_group = SelectionGroup()

        if len(selection_group) >= 1:
            # load the active selection
            self._create_active_selection()
            self._active_selection.selection_box = selection_group[-1]

        if len(selection_group) >= 2:
            # load the selection group
            self._selection.selection_group = selection_group[:-1]

    def _get_editing_selection(self) -> Tuple[NPVector3, NPVector3]:
        """Get the minimum and maximum points of the editing selection.
        This is based on the stored starting pointer and the current pointer."""
        start_point_1 = self._start_point_1.copy()
        start_point_2 = self._start_point_2.copy()
        pointer = self._pointer.bounds.copy()
        start_point_1[:, self._pointer_mask[0]] = pointer[:, self._pointer_mask[0]]
        start_point_2[:, self._pointer_mask[1]] = pointer[:, self._pointer_mask[1]]
        start_1, start_2 = start_point_1
        point_1, point_2 = start_point_2
        start = start_1
        point = point_1
        mask = numpy.abs(start - point_1) < numpy.abs(start - point_2)
        point[mask] = point_2[mask]
        mask = numpy.abs(point - start_1) < numpy.abs(point - start_2)
        start[mask] = start_2[mask]
        return start, point

    def _update_pointer(self):
        # find the closest box position
        # check if there are any blocks up to that position

        # find the closest box position
        # find the closest box or block position
        if self._pointer_moved:
            if self.canvas.camera.projection_mode == Projection.TOP_DOWN:
                camera = self.canvas.camera.location
                camera = (camera[0], 10 ** 9, camera[2])
                look_vector = self.look_vector()
                selection_group, box_index, max_distance = self._get_box_hit_data(
                    camera, look_vector
                )
                location, hit_block = self.closest_block_2d(
                    int(10 ** 9 - min(max_distance, 10 ** 9)) + 1
                )
            else:
                (
                    camera,
                    look_vector,
                    selection_group,
                    box_index,
                    max_distance,
                ) = self._get_default_box_hit_data()
                if self.canvas.camera.rotating:
                    if self._resizing:
                        location = self.distance_block_3d(self._pointer_distance2)
                    else:
                        location = self.distance_block_3d(self._pointer_distance)
                    hit_block = True
                else:
                    if self._resizing:
                        location = self.distance_block_3d(self._pointer_distance2)
                        hit_block = True
                    else:
                        location, hit_block = self.closest_block_3d(
                            min(max_distance, 100)
                        )

            self._pointer.point1, self._pointer.point2 = location, location + 1
            self._pointer_moved = False
            if self._editing:
                (
                    self._active_selection.point1,
                    self._active_selection.point2,
                ) = self._get_editing_selection()
                self._post_change_event()
            else:
                self._selection.reset_highlight_edges()
                if box_index is None or hit_block:
                    # if no box was hit or a block was hit
                    self._highlight = False
                    if self._active_selection is not None:
                        self._active_selection.reset_highlight_edges()
                else:

                    self._highlight = True
                    faces_hit = self._get_box_faces_manual(
                        camera, look_vector, selection_group, box_index, max_distance
                    )
                    location = numpy.floor(
                        camera + look_vector * (max_distance + 0.01)
                    ).astype(int)
                    self._pointer.point1, self._pointer.point2 = location, location + 1
                    if box_index == len(self._selection):
                        self._active_selection.set_highlight_edges(faces_hit)
                    else:
                        self._selection.set_highlight_edges(box_index, faces_hit)
                        self._active_selection.reset_highlight_edges()

    def _get_default_box_hit_data(
        self,
    ) -> Tuple[PointCoordinates, NPVector3, SelectionGroup, Optional[int], float]:
        camera = self.canvas.camera.location
        look_vector = self.look_vector()
        return (camera, look_vector) + self._get_box_hit_data(camera, look_vector)

    def _get_box_hit_data(
        self, camera: PointCoordinates, look_vector: NPVector3
    ) -> Tuple[SelectionGroup, Optional[int], float]:
        selection_group = self.selection_group
        box_index, max_distance = selection_group.closest_vector_intersection(
            camera, look_vector
        )
        return selection_group, box_index, max_distance

    def _get_box_faces(self) -> Tuple[Optional[int], float, NPArray2x3]:
        """Get collision information between a vector and a selection group.
        If the vector intersects a box in the selection group it will return
            The index of the selection box in the selection group.
            A 2x3 bool array of the faces that were hit as corresponding to point1 and point2 x,y,z
        If the vector does not hit the selection the same structure applies but will look like this
            None
            2x3 False array"""
        (
            camera,
            look_vector,
            selection_group,
            box_index,
            max_distance,
        ) = self._get_default_box_hit_data()
        return (
            box_index,
            max_distance,
            self._get_box_faces_manual(
                camera, look_vector, selection_group, box_index, max_distance
            ),
        )

    @staticmethod
    def _get_box_faces_manual(
        camera: PointCoordinates,
        look_vector: NPVector3,
        selection_group: SelectionGroup,
        box_index: Optional[int],
        max_distance: float,
    ) -> NPArray2x3:
        """Get a 2x3 bool array of the faces the look vector first contacts.
        point1 (x y z), point2 (x y z)"""
        if box_index is None:
            # it doesn't hit any boxes
            return numpy.zeros((2, 3), dtype=bool)
        else:
            box: SelectionBox = selection_group[box_index]
            point = max_distance * look_vector + camera
            tol = numpy.min([numpy.array(box.shape) / 4, numpy.ones(3)], 0)
            return numpy.abs(box.points_array - point) < tol

    def draw(self):
        if self.canvas.camera.projection_mode == Projection.TOP_DOWN:
            camera = None
        else:
            camera = self.canvas.camera.location
        self._selection.draw(self.canvas.camera.transformation_matrix, camera)
        if self._active_selection is not None:
            self._active_selection.draw(
                self.canvas.camera.transformation_matrix, camera
            )
        if not self._highlight and not self._editing:
            super().draw()



from amulet_map_editor.api.datatypes import MenuData


class BasePageUI:
    def disable(self):
        pass

    def enable(self):
        pass

    def close(self):
        pass

    def menu(self, menu: MenuData) -> MenuData:
        return menu



import math
import numpy
from amulet.utils.matrix import (
    scale_matrix,
    rotation_matrix_xy,
    rotation_matrix_yx,
    rotation_matrix_xyz,
    transform_matrix,
    displacement_matrix,
    inverse_transform_matrix,
)

TransformationMatrixType = numpy.ndarray


def perspective_matrix(
    fovy: float,  # field of view in the vertical direction in radians
    aspect,  # aspect ratio x/y
    z_near,  # near clipping distance
    z_far,  # far clipping distance
) -> TransformationMatrixType:
    # camera projection
    f = 1 / math.tan(fovy / 2)
    return numpy.array(
        [
            [f / aspect, 0, 0, 0],
            [0, f, 0, 0],
            [
                0,
                0,
                (z_far + z_near) / (z_near - z_far),
                (2 * z_far * z_near) / (z_near - z_far),
            ],
            [0, 0, -1, 0],
        ],
        dtype=numpy.float64,
    )


def orthographic_matrix(
    radius: float,  # the viewable distance in the vertical direction in blocks
    aspect,  # aspect ratio x/y
    near,  # near clipping distance
    far,  # far clipping distance
) -> TransformationMatrixType:
    """The orthographic matrix to convert camera space to screen space."""
    bottom = -radius
    top = radius
    left = bottom * aspect
    right = top * aspect
    return numpy.array(
        [
            [2 / (right - left), 0, 0, -(right + left) / (right - left)],
            [0, 2 / (top - bottom), 0, -(top + bottom) / (top - bottom)],
            [0, 0, -2 / (far - near), -(far + near) / (far - near)],
            [0, 0, 0, 1],
        ]
    )



import wx
from typing import TYPE_CHECKING
from OpenGL.GL import (
    glClear,
    GL_DEPTH_BUFFER_BIT,
)
import traceback
import os

import amulet
from amulet.api.errors import LoaderNoneMatched

from amulet_map_editor import log
from amulet_map_editor.api.wx.ui.traceback_dialog import TracebackDialog
from amulet_map_editor.api.opengl.camera import Projection
from .base_tool_ui import BaseToolUI
from amulet_map_editor.programs.edit.api.events import EVT_DRAW
from amulet_map_editor.programs.edit.api.behaviour import CameraBehaviour

if TYPE_CHECKING:
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas


class DefaultBaseToolUI(BaseToolUI):
    """A base class for tools that adds a render loop, camera behaviour and file dropping logic."""

    def __init__(self, canvas: "EditCanvas"):
        super().__init__(canvas)
        self._camera_behaviour = CameraBehaviour(self.canvas)

    @property
    def name(self) -> str:
        """The name of the tool."""
        raise NotImplementedError

    def enable(self):
        """Set the state of the tool for being enabled."""
        self.canvas.camera.projection_mode = Projection.PERSPECTIVE

    def bind_events(self):
        """Bind all required events to the canvas.
        All events on the canvas will be automatically removed after the tool is disabled.
        """
        self.canvas.Bind(EVT_DRAW, self._on_draw)
        self.canvas.SetDropTarget(None)  # fixes #239
        self.canvas.DragAcceptFiles(True)
        self.canvas.Bind(wx.EVT_DROP_FILES, self._on_drop_files)
        self._camera_behaviour.bind_events()

    def _on_draw(self, evt):
        """The render loop for this tool."""
        self.canvas.renderer.start_draw()
        if self.canvas.camera.projection_mode == Projection.PERSPECTIVE:
            self.canvas.renderer.draw_sky_box()
            glClear(GL_DEPTH_BUFFER_BIT)
        self.canvas.renderer.draw_level()
        self.canvas.renderer.end_draw()

    def _on_drop_files(self, evt: wx.DropFilesEvent):
        """Logic to run when a file is dropped into the canvas."""
        paths = evt.GetFiles()
        if paths:
            pathname = paths[0]
            if os.path.isfile(pathname):
                # TODO: if importing worlds get supported fix this
                try:
                    level = amulet.load_level(pathname)
                except LoaderNoneMatched:
                    msg = f"Could not find a matching loader for {pathname}."
                    wx.MessageBox(msg)
                    log.error(msg)
                except Exception as e:
                    log.error(f"Could not open {pathname}.", exc_info=True)
                    dialog = TracebackDialog(
                        self.canvas,
                        f"Could not open {pathname}.",
                        str(e),
                        traceback.format_exc(),
                    )
                    dialog.ShowModal()
                    dialog.Destroy()
                else:
                    self.canvas.paste(level, level.dimensions[0])
        evt.Skip()



from typing import TYPE_CHECKING, Dict, Tuple, Optional
import wx
from OpenGL.GL import (
    glClear,
    GL_DEPTH_BUFFER_BIT,
    GL_DEPTH_TEST,
    glEnable,
    glDisable,
    glGetBoolean,
)

from amulet_map_editor import lang
from amulet_map_editor.api.opengl.camera import Projection, EVT_CAMERA_MOVED
from amulet_map_editor.programs.edit.api.ui.tool import DefaultBaseToolUI
from amulet_map_editor.programs.edit.api.behaviour import ChunkSelectionBehaviour
from amulet.operations.delete_chunk import delete_chunk
from amulet.api.data_types import Dimension
from amulet_map_editor.programs.edit.plugins.operations.stock_plugins.internal_operations.prune_chunks import (
    prune_chunks,
)

if TYPE_CHECKING:
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas


class ChunkTool(wx.BoxSizer, DefaultBaseToolUI):
    def __init__(self, canvas: "EditCanvas"):
        wx.BoxSizer.__init__(self, wx.HORIZONTAL)
        DefaultBaseToolUI.__init__(self, canvas)

        self._selection = ChunkSelectionBehaviour(self.canvas)

        self._button_panel = wx.Panel(canvas)
        button_sizer = wx.BoxSizer(wx.VERTICAL)
        self._button_panel.SetSizer(button_sizer)

        y_sizer = wx.FlexGridSizer(2, 5, 5)
        button_sizer.Add(y_sizer, flag=wx.ALL, border=5)
        min_y_label = wx.StaticText(
            self._button_panel, label=lang.get("program_3d_edit.chunk_tool.min_y")
        )
        y_sizer.Add(min_y_label, flag=wx.ALIGN_CENTER)
        self._min_y = wx.SpinCtrl(
            self._button_panel,
            min=-30_000_000,
            max=30_000_000,
            initial=256,
        )
        self._min_y.SetToolTip(lang.get("program_3d_edit.chunk_tool.min_y_tooltip"))
        y_sizer.Add(self._min_y, flag=wx.ALIGN_CENTER)
        self._min_y.Bind(wx.EVT_SPINCTRL, self._on_update_clipping)

        max_y_label = wx.StaticText(
            self._button_panel, label=lang.get("program_3d_edit.chunk_tool.max_y")
        )
        y_sizer.Add(max_y_label, flag=wx.ALIGN_CENTER)
        self._max_y = wx.SpinCtrl(
            self._button_panel,
            min=-30_000_000,
            max=30_000_000,
            initial=0,
        )
        self._max_y.SetToolTip(lang.get("program_3d_edit.chunk_tool.max_y_tooltip"))
        y_sizer.Add(self._max_y, flag=wx.ALIGN_CENTER)
        self._max_y.Bind(wx.EVT_SPINCTRL, self._on_update_clipping)
        self._dimensions: Dict[Dimension, Tuple[int, int]] = {}

        delete_button = wx.Button(
            self._button_panel,
            label=lang.get("program_3d_edit.chunk_tool.delete_chunks"),
        )
        delete_button.SetToolTip(
            lang.get("program_3d_edit.chunk_tool.delete_chunks_tooltip")
        )
        button_sizer.Add(
            delete_button, 0, wx.LEFT | wx.BOTTOM | wx.RIGHT | wx.EXPAND, 5
        )
        delete_button.Bind(wx.EVT_BUTTON, self._delete_chunks)

        prune_button = wx.Button(
            self._button_panel,
            label=lang.get("program_3d_edit.chunk_tool.prune_chunks"),
        )
        prune_button.SetToolTip(
            lang.get("program_3d_edit.chunk_tool.prune_chunks_tooltip")
        )
        button_sizer.Add(prune_button, 0, wx.LEFT | wx.BOTTOM | wx.RIGHT | wx.EXPAND, 5)
        prune_button.Bind(wx.EVT_BUTTON, self._prune_chunks)

        self.Add(self._button_panel, 0, wx.ALIGN_CENTER_VERTICAL)

    @property
    def name(self) -> str:
        return "Chunk"

    def bind_events(self):
        super().bind_events()
        self._selection.bind_events()
        self.canvas.Bind(EVT_CAMERA_MOVED, self._on_update_clipping)

    def enable(self):
        self.Layout()
        self.canvas.camera.projection_mode = Projection.TOP_DOWN
        self._selection.enable()
        self._update_clipping()

        dimension = self.canvas.dimension
        if dimension not in self._dimensions:
            self._dimensions[dimension] = (
                min(
                    30_000_000,
                    max(-30_000_000, self.canvas.world.bounds(dimension).min[1]),
                ),
                min(
                    30_000_000,
                    max(-30_000_000, self.canvas.world.bounds(dimension).max[1]),
                ),
            )
        miny, maxy = self._dimensions[dimension]
        self._min_y.SetValue(miny)
        self._max_y.SetValue(maxy)
        self._update_clipping()

    def disable(self):
        super().disable()
        self.canvas.camera.orthographic_clipping = -(10 ** 5), 10 ** 5

    def _on_update_clipping(self, evt):
        self._update_clipping()
        evt.Skip()

    def _update_clipping(self):
        y = self.canvas.camera.location[1]
        self.canvas.camera.orthographic_clipping = (
            y - self._max_y.GetValue() - 1,
            y - self._min_y.GetValue() + 1,
        )

    def _ask_delete_chunks(self) -> Optional[bool]:
        class DeleteChunksDialog(wx.Dialog):
            def __init__(self, *args, **kwds):
                kwds["style"] = kwds.get("style", 0) | wx.DEFAULT_DIALOG_STYLE
                wx.Dialog.__init__(self, *args, **kwds)
                self.SetTitle("Do you want to load the original chunk state?")

                sizer_1 = wx.BoxSizer(wx.VERTICAL)

                label_1 = wx.StaticText(
                    self,
                    wx.ID_ANY,
                    "Do you want to load the original chunk state?\n\n"
                    'Clicking "Yes" will allow you to undo this operation but the operation will take a while to process.\n\n'
                    'Clicking "No" will mean this operation cannot be undone.\n\n'
                    "Changes will not be made to the world until you save so closing before saving will not actually delete the chunks.",
                    style=wx.ALIGN_CENTER_HORIZONTAL,
                )
                label_1.Wrap(500)
                sizer_1.Add(label_1, 0, wx.ALL, 5)

                sizer_2 = wx.StdDialogButtonSizer()
                sizer_1.Add(sizer_2, 0, wx.ALIGN_RIGHT | wx.ALL, 4)

                self.button_YES = wx.Button(self, wx.ID_YES, "")
                self.button_YES.SetDefault()
                sizer_2.AddButton(self.button_YES)

                self.button_NO = wx.Button(self, wx.ID_NO, "")
                self.button_NO.Bind(wx.EVT_BUTTON, self._on_no)
                sizer_2.AddButton(self.button_NO)

                self.button_CANCEL = wx.Button(self, wx.ID_CANCEL, "")
                sizer_2.AddButton(self.button_CANCEL)

                sizer_2.Realize()

                self.SetSizer(sizer_1)
                sizer_1.Fit(self)

                self.SetAffirmativeId(self.button_YES.GetId())
                self.SetEscapeId(self.button_CANCEL.GetId())

                self.Layout()

            def _on_no(self, evt):
                self.EndModal(wx.ID_NO)

        d = DeleteChunksDialog(self.canvas)
        response = d.ShowModal()
        if response == wx.ID_YES:
            return True
        elif response == wx.ID_NO:
            return False
        return None

    def _delete_chunks(self, evt):
        load_original = self._ask_delete_chunks()
        if load_original is not None:
            self.canvas.run_operation(
                lambda: delete_chunk(
                    self.canvas.world,
                    self.canvas.dimension,
                    self.canvas.selection.selection_group,
                    load_original,
                )
            )

    def _prune_chunks(self, evt):
        load_original = self._ask_delete_chunks()
        if load_original is not None:
            self.canvas.run_operation(
                lambda: prune_chunks(
                    self.canvas.world,
                    self.canvas.dimension,
                    self.canvas.selection.selection_group,
                    load_original,
                )
            )

    def _on_draw(self, evt):
        self.canvas.renderer.start_draw()
        if self.canvas.camera.projection_mode == Projection.PERSPECTIVE:
            self.canvas.renderer.draw_sky_box()
            glClear(GL_DEPTH_BUFFER_BIT)
        self.canvas.renderer.draw_level()
        if self.canvas.camera.projection_mode == Projection.PERSPECTIVE:
            self._selection.draw()
        else:
            depth_state = glGetBoolean(GL_DEPTH_TEST)
            if depth_state:
                glDisable(GL_DEPTH_TEST)
            clip = self.canvas.camera.orthographic_clipping
            self.canvas.camera.orthographic_clipping = -(10 ** 5), 10 ** 5
            self._selection.draw()
            self.canvas.camera.orthographic_clipping = clip
            if depth_state:
                glEnable(GL_DEPTH_TEST)
        self.canvas.renderer.end_draw()



import pkgutil

from amulet_map_editor import api, programs

hiddenimports = [
    name for _, name, _ in pkgutil.walk_packages(api.__path__, api.__name__ + ".")
] + [
    name
    for _, name, _ in pkgutil.walk_packages(programs.__path__, programs.__name__ + ".")
]



import numpy
from typing import Tuple, Dict, List, Union

import minecraft_model_reader
from amulet.api.chunk import Chunk

from amulet_map_editor.api.opengl.mesh import TriMesh
from amulet_map_editor.api.opengl.resource_pack import (
    OpenGLResourcePackManagerStatic,
    OpenGLResourcePack,
)

_brightness_step = 0.15
_brightness_multiplier = {
    None: (1,) * 3,
    "up": (1,) * 3,
    "north": (1 - _brightness_step,) * 3,
    "south": (1 - _brightness_step,) * 3,
    "east": (1 - _brightness_step * 2,) * 3,
    "west": (1 - _brightness_step * 2,) * 3,
    "down": (1 - _brightness_step * 3,) * 3,
}


def create_lod0_array(
    larger_blocks: numpy.ndarray,
    unique_blocks: numpy.ndarray,
    models: Dict[int, minecraft_model_reader.BlockMesh],
    texture_bounds: Dict[str, Tuple[float, float, float, float]],
    vert_len: int,
    chunk_offset: Union[Tuple[int, int, int], numpy.ndarray],
    sub_chunk_offset: Tuple[int, int, int] = None,
) -> Tuple[List[numpy.ndarray], List[numpy.ndarray]]:
    """Create chunk geometry for a given sub-chunk.
    Creates a numpy array for opaque geometry and a numpy array for translucent geometry.

    :param larger_blocks: A numpy array for the sub-chunk that extends by one block in each direction.
    :param unique_blocks: A numpy array of the unique block ids in the sub-chunk.
    :param models: A dictionary mapping internal block id to the block model.
    :param texture_bounds: A dictionary mapping texture id to the bounds of the texture in the atlas.
    :param vert_len: The number of float32 values per vertex.
    :param chunk_offset: The offset of the chunk in the world.
    :param sub_chunk_offset: The offset of the sub-chunk relative to the chunk.
    :return:
    """
    sub_chunk_offset = sub_chunk_offset or (0, 0, 0)

    blocks = larger_blocks[1:-1, 1:-1, 1:-1]

    transparent_array = numpy.zeros(larger_blocks.shape, dtype=numpy.uint8)
    for block_temp_id in unique_blocks:
        model = models[block_temp_id]
        transparent_array[larger_blocks == block_temp_id] = model.is_transparent

    def get_transparent_array(offset_transparent_array, transparent_array_):
        return numpy.logical_and(
            offset_transparent_array,  # if the next block is transparent
            numpy.logical_not(  # but is not the same block with transparency mode 1
                (offset_transparent_array == 1)
                * (offset_transparent_array == transparent_array_)
            ),
        )

    middle_transparent_array = transparent_array[1:-1, 1:-1, 1:-1]
    show_up = get_transparent_array(
        transparent_array[1:-1, 2:, 1:-1], middle_transparent_array
    )
    show_down = get_transparent_array(
        transparent_array[1:-1, :-2, 1:-1], middle_transparent_array
    )
    show_north = get_transparent_array(
        transparent_array[1:-1, 1:-1, :-2], middle_transparent_array
    )
    show_south = get_transparent_array(
        transparent_array[1:-1, 1:-1, 2:], middle_transparent_array
    )
    show_east = get_transparent_array(
        transparent_array[2:, 1:-1, 1:-1], middle_transparent_array
    )
    show_west = get_transparent_array(
        transparent_array[:-2, 1:-1, 1:-1], middle_transparent_array
    )

    show_map = {
        "up": show_up,
        "down": show_down,
        "north": show_north,
        "south": show_south,
        "east": show_east,
        "west": show_west,
    }

    chunk_verts = []
    chunk_verts_translucent = []

    x, y, z = blocks.shape
    block_locations = numpy.transpose(numpy.mgrid[0:x, 0:y, 0:z], (1, 2, 3, 0)).reshape(
        (-1, 3)
    )
    unique_blocks, inverse, counts = numpy.unique(
        blocks, return_inverse=True, return_counts=True
    )
    block_indexes = numpy.argsort(inverse)
    block_locations = numpy.split(
        block_locations[block_indexes], numpy.cumsum(counts)[:-1]
    )

    for block_temp_id, all_block_locations in zip(unique_blocks, block_locations):
        model = models[block_temp_id]
        # for each unique blockstate in the chunk
        # get the model and the locations of the blocks
        # if not all_block_locations.size:
        #     continue
        where = None
        for cull_dir in model.faces.keys():
            # iterate through each cull direction
            # narrow down the blocks to what should be created for that cull direction
            if cull_dir is None:
                block_locations = all_block_locations
            elif cull_dir in show_map:
                if where is None:
                    where = tuple(all_block_locations.T)
                block_locations = all_block_locations[show_map[cull_dir][where]]
                if not block_locations.size:
                    continue
            else:
                continue

            # the number of blocks and their offsets in chunk space
            block_count = len(block_locations)
            block_offsets = block_locations

            # the vertices in model space
            verts = model.verts[cull_dir].reshape((-1, 3))
            tverts = model.texture_coords[cull_dir].reshape((-1, 2))
            faces = model.faces[cull_dir]

            # each slice in the first axis is a new block, each slice in the second is a new vertex
            vert_table = numpy.zeros(
                (block_count, faces.size, vert_len), dtype=numpy.float32
            )
            vert_table[:, :, :3] = (
                verts[faces]
                + block_offsets[:, :].reshape((-1, 1, 3))
                + chunk_offset
                + sub_chunk_offset
            )
            vert_table[:, :, 3:5] = tverts[faces]

            vert_index = 0
            for texture_index in model.texture_index[cull_dir]:
                tex_bounds = texture_bounds[model.textures[texture_index]]

                vert_table[:, vert_index : vert_index + 3, 5:9] = tex_bounds
                vert_index += 3

            vert_table[:, :, 9:12] = (
                model.tint_verts[cull_dir].reshape((-1, 3))[faces]
                * _brightness_multiplier[cull_dir]
            )
            vert_table[:, :, 9:12] *= 0.9 + 0.2 * numpy.abs(
                (numpy.remainder(vert_table[:, :, 1:2] / 32, 2) - 1)
            )

            if model.is_transparent == 1:
                chunk_verts_translucent.append(vert_table.ravel())
            else:
                chunk_verts.append(vert_table.ravel())

    return chunk_verts, chunk_verts_translucent


class RenderChunkBuilder(TriMesh, OpenGLResourcePackManagerStatic):
    """A class to define the logic to generate geometry from a block array"""

    def __init__(self, context_identifier: str, resource_pack: OpenGLResourcePack):
        texture = resource_pack.get_atlas_id(context_identifier)
        TriMesh.__init__(self, context_identifier, texture)
        OpenGLResourcePackManagerStatic.__init__(self, resource_pack)

    @property
    def chunk(self) -> Chunk:
        raise NotImplementedError

    @property
    def offset(self) -> numpy.ndarray:
        raise NotImplementedError

    @staticmethod
    def _get_block_data(blocks: numpy.ndarray) -> Tuple[numpy.ndarray, numpy.ndarray]:
        """Given a Chunk object will return the chunk arrays needed to generate geometry
        :returns: block array of the chunk, block array one block larger than the chunk, array of unique blocks"""
        larger_blocks = numpy.zeros(blocks.shape + numpy.array((2, 2, 2)), blocks.dtype)
        larger_blocks[1:-1, 1:-1, 1:-1] = blocks
        unique_blocks = numpy.unique(larger_blocks)
        return larger_blocks, unique_blocks

    def create_geometry(self):
        raise NotImplementedError

    def _set_verts(
        self,
        chunk_verts: List[numpy.ndarray],
        chunk_verts_translucent: List[numpy.ndarray],
    ):
        if chunk_verts:
            self.verts = numpy.concatenate(chunk_verts, 0)
            self.verts_translucent = self.verts.size
        else:
            self.verts = self.new_empty_verts()

        if chunk_verts_translucent:
            chunk_verts_translucent.insert(0, self.verts)
            self.verts = numpy.concatenate(chunk_verts_translucent, 0)

        self.draw_count = int(self.verts.size // self._vert_len)

    def _create_lod0_multi(
        self, blocks: List[Tuple[numpy.ndarray, Tuple[int, int, int]]]
    ) -> Tuple[List[numpy.ndarray], List[numpy.ndarray]]:
        """Create LOD0 geometry data for every sub-chunk in a given chunk.

        :param blocks: A list of tuples containing block arrays extending one block outside the sub-chunk in each direction.
        :return: Opaque block vertices, translucent block vertices.
        """
        chunk_verts = []
        chunk_verts_translucent = []

        if blocks:
            # unique blocks per sub-chunk
            unique_sub_chunk_blocks = [numpy.unique(arr) for arr, _ in blocks]

            # unique blocks in the chunk
            unique_chunk_blocks = numpy.unique(
                numpy.concatenate(unique_sub_chunk_blocks)
            )

            block_palette = self.chunk.block_palette
            models: Dict[int, minecraft_model_reader.BlockMesh] = {
                block_temp_id: self.resource_pack.get_block_model(
                    block_palette[block_temp_id]
                )
                for block_temp_id in unique_chunk_blocks
            }
            texture_bounds: Dict[str, Tuple[float, float, float, float]] = {
                texture_path: self.resource_pack.texture_bounds(texture_path)
                for model in models.values()
                for texture_path in model.textures
            }

            for (larger_blocks, offset), unique_blocks in zip(
                blocks, unique_sub_chunk_blocks
            ):
                chunk_verts_, chunk_verts_translucent_ = create_lod0_array(
                    larger_blocks,
                    unique_blocks,
                    models,
                    texture_bounds,
                    self._vert_len,
                    self.offset,
                    offset,
                )
                chunk_verts += chunk_verts_
                chunk_verts_translucent += chunk_verts_translucent_

        return chunk_verts, chunk_verts_translucent
